/*
* ==========================================================================
*       Filename:  antzer_faro_can_j1708.c
*    Description:  Get/Send raw can and J1708 data.
*        Version:  1.0
*        Created:  2020.12.31
*       Compiler:  gcc
*         Author:  Vincent
*        Company:  Antzer
* ==========================================================================
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>
#include <time.h>
#include "faro_can_sdk.h"

#define CAN_P0_ID 0x18FEC100
#define CAN_P1_ID 0x0CF00400
#define J1708_MID 0x86
#define CAN_P0_SEND_PERIOD 50000 //50ms
#define CAN_P1_SEND_PERIOD 100000 //100ms
#define J1708_SEND_PERIOD 600000 //600ms

#define J1708_21B_DATA_MAX 13

pthread_mutex_t bus_mutex;
pthread_t do_read_CAN_J1708_thread;
pthread_t do_send_CAN_p0_thread;
pthread_t do_send_CAN_p1_thread;
pthread_t do_send_J1708_thread;
int stop_thread = 0;

//CAN port 0
unsigned long sendCAN_p0_cnt = 0; 
unsigned long recCAN_p0_cnt = 0;
//CAN port 1
unsigned long sendCAN_p1_cnt = 0;
unsigned long recCAN_p1_cnt = 0;
//J1708
unsigned long sendJ1708_cnt = 0;
unsigned long recJ1708_cnt = 0;

//J1708 data
struct J1708_message_t J1708_msg_keep;
int read_J1708_msg_21_no = 1; //1: first package, 2: second package

//================Read===============
//-------------------------
int read_J1708_msg(struct J1708_message_t* msg) //HAL
{
	int j1708_amount;
	int j;
	int read_size = 0;
	int ret;
	
	struct J1708_21_message_t read_J1708_msg_21;
//    pthread_mutex_lock(&bus_mutex);
	ret = AZ_VC_J1708_21_Read(&read_J1708_msg_21, &read_size);
//	pthread_mutex_unlock(&bus_mutex);
	if(ret == 0) {

		if((read_J1708_msg_21.pkt_no == 1) && (read_J1708_msg_21.dlc <= J1708_21B_DATA_MAX)) {
			msg->mid = read_J1708_msg_21.mid;
			msg->p = read_J1708_msg_21.p;
			msg->dlc = read_J1708_msg_21.dlc;
			for(j = 0; j < read_J1708_msg_21.dlc; j++) {
				msg->data[j] = read_J1708_msg_21.data[j];
			}
			read_J1708_msg_21_no = 1;
			return 0; //Got data
		}
		else
		if((read_J1708_msg_21.pkt_no == 1) && (read_J1708_msg_21.dlc > J1708_21B_DATA_MAX)) {
			J1708_msg_keep.mid = read_J1708_msg_21.mid;
			J1708_msg_keep.p = read_J1708_msg_21.p;
			J1708_msg_keep.dlc = read_J1708_msg_21.dlc;
			for(j = 0; j < J1708_21B_DATA_MAX; j++) {
				J1708_msg_keep.data[j] = read_J1708_msg_21.data[j];
			}
			read_J1708_msg_21_no = 2;
			return 1; //Wait second package
		}
		else
		if((read_J1708_msg_21.pkt_no == 2) && (read_J1708_msg_21_no == 2)) {
			for(j = 0; j < (J1708_msg_keep.dlc - J1708_21B_DATA_MAX); j++) {
				J1708_msg_keep.data[(J1708_21B_DATA_MAX + j)] = read_J1708_msg_21.data[j];
			}
			read_J1708_msg_21_no = 1;
			return 0; //Got data
		}
		else
		{
			read_J1708_msg_21_no = 1;
		}
	}
	return -1; //Error
}

//-------------------------
void get_J1708_data() {
	int i;
    int dataLeng;
	int ret = 0;
	struct J1708_message_t J1708_msg;
	uint32_t size = 0;
//    pthread_mutex_lock(&bus_mutex);
    	dataLeng = AZ_VC_J1708_21_Read_Amount();
//	pthread_mutex_unlock(&bus_mutex);
	for(i = 0; i < dataLeng; i++) //The amount of can data received.
	{
		ret = read_J1708_msg(&J1708_msg);
		if(ret == 0) { 
			recJ1708_cnt++;
			printf("rec J1708 data Cnt:%ld, MID: ID:%X, DLC:%x\n", recJ1708_cnt, J1708_msg.mid , J1708_msg.dlc);
		}
		else if(ret == -1)
		{
			printf("ERROR:Failed to Read J1708 message.\n");
		}
	}
}

//-------------------------
void get_can_data() {
	int i;
    int dataLeng;
	int ret;
	struct can_message_t msg;
	uint32_t size = 0;
	memset(&msg, 0x00, sizeof(msg));
	size = 0;
//    pthread_mutex_lock(&bus_mutex);
    dataLeng = AZ_VC_CAN_Read_Amount();
//	pthread_mutex_unlock(&bus_mutex);
	for(i = 0; i < dataLeng; i++) //The amount of can data received.
	{
//		pthread_mutex_lock(&bus_mutex);
		ret = AZ_VC_CAN_Read(&msg, &size);
//		pthread_mutex_unlock(&bus_mutex);
		if(ret == 0) { 
			if(msg.port == 0) {
				recCAN_p0_cnt++;
				printf("rec CAN Frame Cnt:%ld, ID:%X, Port:%x\n", recCAN_p0_cnt,msg.id, msg.port);
			}
			else {
				recCAN_p1_cnt++;
				printf("rec CAN Frame Cnt:%ld, ID:%X, Port:%x\n", recCAN_p1_cnt,msg.id, msg.port);
			}
			
			printf("CAN Data : %02x %02x %02x %02x %02x %02x %02x %02x\n",msg.data[0], msg.data[1], msg.data[2], msg.data[3], msg.data[4], msg.data[5], msg.data[6], msg.data[7]);
		}
		else
		{
			printf("ERROR:Failed to Read message from CAN Port.\n");
		}
	}
	
}

//-------------------------
static void *read_can_j1708(void *args)
{
	recJ1708_cnt = 0;
	recCAN_p0_cnt = 0;
	recCAN_p1_cnt = 0;
	while(stop_thread != 1)
	{
		get_can_data();
		get_J1708_data();
		usleep(1000); //1ms
	}
	pthread_exit(NULL);
}

//================Send===============
//-------------------------
int send_J1708_msg(struct J1708_message_t msg) //HAL
{
	int i;
	int ret;
	int len;
	struct J1708_21_message_t send_J1708_msg_21;

	if(msg.dlc > J1708_21B_DATA_MAX) {
		len = J1708_21B_DATA_MAX;
	}
	else {
		len = msg.dlc;
	}

	send_J1708_msg_21.mid = msg.mid;
	send_J1708_msg_21.p = msg.p; //Priority
	send_J1708_msg_21.dlc = msg.dlc;
	send_J1708_msg_21.pkt_no = 1; //1:first package, 2: second package

	for(i = 0; i < len; i++)
	{
		send_J1708_msg_21.data[i] = msg.data[i];
	}
//	pthread_mutex_lock(&bus_mutex);	
	ret = AZ_VC_J1708_21_Write(send_J1708_msg_21);
//	pthread_mutex_unlock(&bus_mutex);
	if(ret != 0)
		return -1;
	
	if(msg.dlc > J1708_21B_DATA_MAX)
	{
		send_J1708_msg_21.pkt_no = 2;

		for(i = 0; i < (msg.dlc - J1708_21B_DATA_MAX); i++)
		{
			send_J1708_msg_21.data[i] = msg.data[(i + J1708_21B_DATA_MAX)];
		}
//		pthread_mutex_lock(&bus_mutex);	
		ret = AZ_VC_J1708_21_Write(send_J1708_msg_21);
//		pthread_mutex_unlock(&bus_mutex);
		if(ret != 0)
			return -1;
	}

	return 0;
}

//-------------------------
void send_J1708_data() {
	int ret;
	int i;
	struct J1708_message_t j1708_msg;
	j1708_msg.mid = J1708_MID;
	j1708_msg.p = 3; //Priority
	j1708_msg.dlc = 19; //Max dlc is 19.
	for(i = 0; i < 19; i++)
	{
		j1708_msg.data[i] = i;
	}
	sendJ1708_cnt++;
	ret = send_J1708_msg(j1708_msg);
	if(ret == 0) {
		printf("send J1708 cnt: %ld, MID:%X\n", sendJ1708_cnt, j1708_msg.mid);
	}
	else {
		printf("ERROR: Failed to send J1708 data");
	}
	pthread_mutex_unlock(&bus_mutex);
}

//-------------------------
void send_CAN_p0_data() {
	int i;
	int amt;
	int ret = 0;
	struct can_message_t msg;
	uint32_t size = 0;
	memset(&msg, 0x00, sizeof(msg));
	size = 0;
	msg.id = CAN_P0_ID;	
	msg.port = 0;
	msg.mode = 1; //0:Standard, 1:Extended
	msg.dlc = 8;
	for(i = 0; i < 8; i++)
	{
		msg.data[i] = i;
	}

	sendCAN_p0_cnt++;
//	pthread_mutex_lock(&bus_mutex);
	ret = AZ_VC_CAN_Write(msg);
//	pthread_mutex_unlock(&bus_mutex);
	if(ret == 0) { 
		printf("send CAN Port0 cnt: %ld, CAN ID:%X\n", sendCAN_p0_cnt, msg.id);
		
	}
	else
	{
		printf("ERROR:Failed to sead message to CAN Port 0.\n");
	}	
}

//-------------------------
void send_CAN_p1_data() {
	int i;
	int amt;
	int ret = 0;
	struct can_message_t msg;
	uint32_t size = 0;
	memset(&msg, 0x00, sizeof(msg));
	size = 0;
	msg.id = CAN_P1_ID;	
	msg.port = 1;
	msg.mode = 1; //0:Standard, 1:Extended
	msg.dlc = 8;
	for(i = 0; i < 8; i++)
	{
		msg.data[i] = i;
	}

	sendCAN_p1_cnt++;
//	pthread_mutex_lock(&bus_mutex);
	ret = AZ_VC_CAN_Write(msg);
//	pthread_mutex_unlock(&bus_mutex);
	if(ret == 0) { 
		printf("send CAN Port1 cnt: %ld, CAN ID:%X\n", sendCAN_p1_cnt, msg.id);
	}
	else
	{
		printf("ERROR:Failed to sead message to CAN Port 1.\n");
	}
}

//-------------------------
static void *send_j1708(void *args)
{
	sendJ1708_cnt = 0;
	while(stop_thread != 1) {
        send_J1708_data();
		usleep(J1708_SEND_PERIOD); //600ms
	}

	pthread_exit(NULL);
}

//-------------------------
static void *send_can_p0(void *args)
{
	sendCAN_p0_cnt = 0; 
	while(stop_thread != 1) {
        send_CAN_p0_data();
		usleep(CAN_P0_SEND_PERIOD); //50ms
	}

	pthread_exit(NULL);
}

//-------------------------
static void *send_can_p1(void *args)
{
	sendCAN_p1_cnt = 0; 
	while(stop_thread != 1) {
        send_CAN_p1_data();
		usleep(CAN_P1_SEND_PERIOD); //100ms
	}

	pthread_exit(NULL);
}

//================Stop===============
//-------------------------
void sig_handler(int signo)
{
	if(signo == SIGINT) {
		fprintf(stderr, "catch SIGINT\n");
		stop_thread = 1;
		printf("CAN port0 send: %ld, rec: %ld\n", sendCAN_p0_cnt, recCAN_p0_cnt);
		printf("CAN port1 send: %ld, rec: %ld\n", sendCAN_p1_cnt, recCAN_p1_cnt);
		printf("J1708 send: %ld, rec: %ld\n", sendJ1708_cnt, recJ1708_cnt);
		exit(0);
	}
}

//================Main===============
//-------------------------
int main()
{
	//char port_name[] = {"/dev/faro_node"};
	char port_name[] = {"/dev/ttyUSB0"};
	int count = 0;
	struct can_fw_ver_t canfwver;
	struct can_config_t canCFG;
	int ret =0;
	read_J1708_msg_21_no = 1; //Check first or second package of J1708 data.
	
	if(signal(SIGINT, sig_handler) == SIG_ERR)
	fprintf(stderr, "fail to catch SIGINT\n");


	// initialize mutex
	pthread_mutex_init(&bus_mutex, NULL);

	if(port_name == NULL) {
		fprintf(stderr, "Invalid port_name, assigned by -p /dev/ttyUSBx\n");
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, 4)) == 0) //921600bps
	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
		}
		else {
			fprintf(stdout, "Get firmware version : v%02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
		}
	}
	else
	{
		fprintf(stderr, "Calling AZ_VC_Init fail\n");	
		return 0;
	}	
	
	/*Activate the mode 0 in CAN port0*/
	ret = AZ_VC_ModeActive(0x00,0x00);
	
	if(ret != 0)
	{
		printf("ERROR: Failed to switch mode");
	}
	
	/*Activate the mode 0 in CAN port1*/
	ret = AZ_VC_ModeActive(0x01,0x00);
	
	if(ret != 0)
	{
		printf("ERROR: Failed to switch mode");
	}
	
	/*Activate the mode J1708 21 bytes*/
	ret = AZ_VC_ModeActive(0x00,0x08); //0x08: J1708 ON; 0x09: J1708 OFF
	
	if(ret != 0)
	{
		printf("ERROR: Failed to switch mode");
	}

	/*Configure the BaudRate*/
	canCFG.port = 0;
	//0x00:1Mbps, 0x01:800Kbps, 0x02:500Kbps, 0x03:250Kbps, 0x04:200 Kbps, 0x05:125 Kbps, 0x08:OFF
	canCFG.speed = 0x02; //500k bps
	if((ret = AZ_VC_CAN_Config(canCFG)) != 0) {
		fprintf(stderr, "Set baud rate of CAN port0 fail\n");
	}
	canCFG.port = 1;
	canCFG.speed = 0x03; //250k bps
	if((ret = AZ_VC_CAN_Config(canCFG)) != 0) {
		fprintf(stderr, "Set baud rate of CAN port1 fail\n");
	}

	if((ret = pthread_create(&do_read_CAN_J1708_thread, NULL, read_can_j1708, NULL)) != 0) {
		fprintf(stderr, "create do_read_CAN_J1708_thread thread fail\n");
	}

	if((ret = pthread_create(&do_send_J1708_thread, NULL, send_j1708, NULL)) != 0) {
		fprintf(stderr, "create do_send_J1708_thread thread fail\n");
	}

	if((ret = pthread_create(&do_send_CAN_p0_thread, NULL, send_can_p0, NULL)) != 0) {
		fprintf(stderr, "create do_send_CAN_p0_thread thread fail\n");
	}
	
	if((ret = pthread_create(&do_send_CAN_p1_thread, NULL, send_can_p1, NULL)) != 0) {
		fprintf(stderr, "create do_send_CAN_p1_thread thread fail\n");
	}

	pthread_join(do_read_CAN_J1708_thread, NULL);
	pthread_join(do_send_J1708_thread, NULL);
	pthread_join(do_send_CAN_p0_thread, NULL);
	pthread_join(do_send_CAN_p1_thread, NULL);
	
	return 0;
}
