/***********************************************************************
 *
 *  Copyright (c) 2019 - 2020 ANTZER TECH CO., LTD.
 *  All Rights Reserved
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <linux/limits.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "faro_can_sdk_demo_config.h"
#include "faro_can_sdk.h"

#define fprintf(fmt,...) {if (!g_quiet_print_flag) printf(__VA_ARGS__);}

//RawCAN
#define FARO_CAN_SDK_RAWCAN_WRITE_ARGS						9
#define FARO_CAN_SDK_RAWCAN_READ_ARGS						1
#define FARO_CAN_SDK_RAWCAN_2PORT_WRITE_ARGS				16
#define FARO_CAN_SDK_RAWCAN_WRITE_READ_ARGS				9
#define FARO_CAN_SDK_RAWCAN_2PORT_WRITE_READ_ARGS		16

#define FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_32_BITS		6
#define FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_16_BITS		8
#define FARO_CAN_SDK_FILTER_CONFIG_RESET_REMOVE			4

//J1939
#define FARO_CAN_SDK_J1939_RW_ARGS						6
#define FARO_CAN_SDK_J1939_PORT_CONFIG_ARGS				2
#define FARO_CAN_SDK_J1939_FILTER_CONFIG_ARGS			4
#define FARO_CAN_SDK_J1939_GET_FILTER_COUNTER_ARGS		2
#define FARO_CAN_SDK_J1939_FILTER_GET_ARGS				2
#define FARO_CAN_SDK_J1939_READ							1

//Can config
#define FARO_CAN_SDK_CAN_CONFIG_ARGS					2
#define FARO_CAN_SDK_GET_FILTER_CONFIG					2

//auto request
#define FARO_CAN_SDK_CAN_REQUEST_ARGS					2

//J1708 21byte
#define FARO_CAN_SDK_DEMO_J1708_READ_21					1
#define FARO_CAN_SDK_DEMO_J1708_WRITE_21				9
#define FARO_CAN_SDK_J1708_FILTER_CONFIG_ARGS			3
#define FARO_CAN_SDK_J1708_GET_FILTER_COUNTS_ARGS		1
#define FARO_CAN_SDK_J1708_FILTER_GET_ARGS				1

//ecompass
#define FARO_CAN_SDK_DEMO_ecompass_set_acu				2
#define FARO_CAN_SDK_DEMO_ecompass_get_acu				1

//Algorithm
#define FARO_CAN_SDK_DEMO_algorithm_clear_gogf			1
#define FARO_CAN_SDK_DEMO_algorithm_set_fb_swh			2
#define FARO_CAN_SDK_DEMO_algorithm_get_fb_swh			1

enum RawCAN_IDX
{
	DEMO_TEST_IDX_RAWCAN_READ = 0,
	DEMO_TEST_IDX_RAWCAN_WRITE,
	DEMO_TEST_IDX_RAWCAN_2PORT_WRITE,
	DEMO_TEST_IDX_RAWCAN_WRITE_READ,
	DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ,
	NUM_OF_DEMO_RAWCAN_TEST_IDX	//must for last one, check Item is correct
};

enum J1939_TEST_IDX
{
	J1939_TEST_IDX_READ = 0,
	J1939_TEST_IDX_RW,
	J1939_TEST_IDX_PORT_CONFIG,
	J1939_TEST_IDX_FILTER_CONFIG,
	J1939_TEST_IDX_GET_FILTER_COUNTER,
	J1939_TEST_IDX_FILTER_GET,
	NUM_OF_DEMO_J1939_TEST_IDX	//must for last one, check Item is correct
};

enum J1708_21_IDX
{
	DEMO_TEST_IDX_J1708_READ_21,
	DEMO_TEST_IDX_J1708_WRITE_21,
	J1708_TEST_IDX_FILTER_CONFIG,
	J1708_TEST_IDX_GET_FILTER_COUNTER,	
	J1708_TEST_IDX_FILTER_GET,
	NUM_OF_DEMO_J1708_TEST_IDX	//must for last one, check Item is correct
};

enum ecompass_demo_IDX
{
	DEMO_TEST_IDX_ecompass_set_accelerator,
	DEMO_TEST_IDX_ecompass_get_accelerator,
	NUM_OF_DEMO_ecompass_IDX	//must for last one, check Item is correct
};

enum Algorithm_demo_IDX
{
	DEMO_TEST_IDX_ALGORITHM_CLEAR_G0Gf = 0,
	DEMO_TEST_IDX_ALGORITHM_SET_F_B_SWH,
	DEMO_TEST_IDX_ALGORITHM_GET_F_B_SWH,
	NUM_OF_DEMO_ALGORITHM_IDX	//must for last one, check Item is correct
};

struct j1939_filter_config_t
{
	uint8_t type;
	uint8_t port;
	uint32_t pgn;
};

struct j1708_21_alldata_t
{
	uint8_t alldata[19];
};

struct J1708_filter_config_t
{	
	uint8_t type;	
	uint8_t mid;
};

//RawCAN
//static struct can_message_t rawcan_send_msg;
static uint8_t rawcan_demo = 0;
static int rawcan_demo_arg = 0;
static struct can_message_t rawcan_send_msg[2];
static uint32_t rawcan_delay_time = 1000;	//for 1000us = 1ms.

static struct can_filter_config_t can_filter_config;
static struct J1939_send_msg_t j1939_send_msg;
static struct j1939_filter_config_t j1939_filter_config;
static struct can_filter_config_t can_filter_config;
static struct J1708_21_message_t j1708_21_send_msg;
static struct module_sku_format_t module_sku;
static struct j1708_21_alldata_t j1708_send_allmsg;
static struct J1708_filter_config_t j1708_filter_config;

static int deinit_in_seconds = (10 * 3);	//set deinit time
static int stop_thread = 0;
static char *port_name = NULL;
static uint8_t get_can_fw_ver = 0;
static uint8_t scan_port = 0;
static uint8_t can_config = 0;
static struct can_config_t can_config_data;
static int test_cnt = 1;
static uint8_t mode_active = 0;
static int mode_active_port = 0;
static int mode_active_arg = 1;
static uint8_t filter_config = 0;
static uint8_t filter_config_raw = 0;
static uint8_t get_filter_config = 0;
static uint8_t read_write_raw = 0;
static uint8_t launch_receiver = 0;
static uint8_t j1939_test = 0;
static int j1939_test_arg = 0;
static int j1939_test_port = 0;
static uint8_t reset_module = 0;
static uint8_t store_config = 0;
static uint8_t store_config_mode = 0;
//For Uart
static uint8_t get_uart_canbaudrate = 0;
static uint8_t set_uart_baudrate = 0;
static uint8_t set_uart_baudrate_arg = 0;

//auto request
static uint8_t set_request = 0;
static uint8_t set_request_time_port = 0;
static uint8_t set_request_time_time = 0;

//J1708_21byte
static uint8_t j1708_21_demo = 0;
static int j1708_21_demo_arg = 0;

//ecompass
static uint8_t ecompass_demo = 0;
static int ecompass_demo_arg = 0;
static int ecompass_acc_scale = 0;

//Algorithm Demo
static uint8_t algorithm_demo = 0;
static int algorithm_demo_arg = 0;
static int algorithm_enable = 0;

//Sku
static uint8_t read_sku = 0;

static void display_help(int argc, char *argv[])
{
	fprintf(stdout, "Usage: %s [OPTION]\n"
			"\n"
			"  -h, --help\n"
			"  -s, --scan, scan available serial port\n"
			"  -p, --port\n"
			"		#1-port, ex \"/dev/ttyUSB0\"\n"
			"  -d, --deinit\n"
			"		#1-deinit after x seconds, -1 endless until ctrl-c\n"
			"  -g, --get_fw_ver, get CAN firmware version\n"
			"  -c, --can_config, can config\n"
			"		#1-port: 1 byte\n"
			"		#2-speed: 1 byte, 0: 1M, 1: 800K, 2: 500K, 3: 250K, 4: 200K, 5: 125K\n"
			"  -b, --reset_module\n"
			"  -m, --mode_active\n"
			"		#1-port: 0 or 1\n"
			"		#2-mode active: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR_OBD2, 5:CAN2ADR_J1939, 8:J1708_21, 9: J1708_21_OFF\n"
			"  -a, --set_request_time\n"
			"		#1-port: 0 or 1\n"
			"		#2-request time: 1~100ms\n"
			"  -f, --filter_config\n"
			"  -w, --filter_config_raw\n"
			"		#1-type: 0: ID+Mask, 1: ID List, 2: Remove, 3: Reset All\n"
			"		#2-port: 0, 1\n"
			"		#3-bank: 0 - 13\n"
			"		#4-mode: 0: 16 bits, 1: 32 bits\n"
			"		#5-filterId: 16/32 bits value in hex, ex: 0x0011/0x00112233\n"
			"		#6-filterId/filterMask: 16bits/32 bits in hex, ex: 0x1234/0x00112233\n"
			"		#7-filterMask: 16 bits in hex, ex: 0x0011\n"
			"		#8-filterMask: 16 bits in hex, ex: 0x0011\n"
			"  -x, --get_filter_config\n"
			"       #1 port: 0, 1\n"
			"       #2 bank: 0 - 13\n"
			"  -r, --RawCan_Demo\n"
			"		0-Read can data\n"
			"		1-Write\n"
			"			#1-set port: 1 byte, value: 0-port0, 1-port1\n"
			"			#2-id: 4 byte, ex: 0x000007DF\n"
			"			#3-can mode: 1 byte, value: 0-2.0A, 1-2.0B\n"
			"			#4-rtr: 1 byte, value: 0-rtr Disable, 1-rtr Enable\n"
			"			#5-dlc: 1 byte, value: 0 ~ 8\n"
			"			#6-4/8 can data(0~3): 4 byte, ex: 0x01020304\n"
			"			#7-4/8 can data(4~7): 4 byte, ex: 0x05060708\n"
			"			#8-delay time after seding: 1 byte, millisecond time, ex:100 -- delay 100ms\n"
			"		2-Write with 2 Data format\n"
			"			#1-(format1)set port: 1 byte, value: 0-port0, 1-port1\n"
			"			#2-(format1)id: 4 byte, ex: 0x000007DF\n"
			"			#3-(format1)can mode: 1 byte, value: 0-2.0A, 1-2.0B\n"
			"			#4-(format1)rtr: 1 byte, value: 0-rtr Disable, 1-rtr Enable\n"
			"			#5-(format1)dlc: 1 byte, value: 0 ~ 8\n"
			"			#6-(format1)4/8 can data(0~3): 4 byte, ex: 0x01020304\n"
			"			#7-(format1)4/8 can data(4~7): 4 byte, ex: 0x05060708\n"
			"			#8-(format2)set port: 1 byte, value: 0-port0, 1-port1\n"
			"			#9-(format2)id: 4 byte, ex: 0x000007DF\n"
			"			#10-(format2)can mode: 1 byte, value: 0-2.0A, 1-2.0B\n"
			"			#11-(format2)rtr: 1 byte, value: 0-rtr Disable, 1-rtr Enable\n"
			"			#12-(format2)dlc: 1 byte, value: 0 ~ 8\n"
			"			#13-(format2)4/8 can data(0~3): 4 byte, ex: 0x01020304\n"
			"			#14-(format2)4/8 can data(4~7): 4 byte, ex: 0x05060708\n"
			"			#15-delay time after seding: 1 byte, millisecond time, ex:100 -- delay 100ms\n"
			"		3-Write and receive\n"
			"			#1-set port: 1 byte, value: 0-port0, 1-port1\n"
			"			#2-id: 4 byte, ex: 0x000007DF\n"
			"			#3-can mode: 1 byte, value: 0-2.0A, 1-2.0B\n"
			"			#4-rtr: 1 byte, value: 0-rtr Disable, 1-rtr Enable\n"
			"			#5-dlc: 1 byte, value: 0 ~ 8\n"
			"			#6-4/8 can data(0~3): 4 byte, ex: 0x01020304\n"
			"			#7-4/8 can data(4~7): 4 byte, ex: 0x05060708\n"
			"			#8-delay time after seding: 1 byte, millisecond time, ex:100 -- delay 100ms\n"
			"		4-Write with 2 Data format and receive\n"
			"			#1-(format1)set port: 1 byte, value: 0-port0, 1-port1\n"
			"			#2-(format1)id: 4 byte, ex: 0x000007DF\n"
			"			#3-(format1)can mode: 1 byte, value: 0-2.0A, 1-2.0B\n"
			"			#4-(format1)rtr: 1 byte, value: 0-rtr Disable, 1-rtr Enable\n"
			"			#5-(format1)dlc: 1 byte, value: 0 ~ 8\n"
			"			#6-(format1)4/8 can data(0~3): 4 byte, ex: 0x01020304\n"
			"			#7-(format1)4/8 can data(4~7): 4 byte, ex: 0x05060708\n"
			"			#8-(format2)set port: 1 byte, value: 0-port0, 1-port1\n"
			"			#9-(format2)id: 4 byte, ex: 0x000007DF\n"
			"			#10-(format2)can mode: 1 byte, value: 0-2.0A, 1-2.0B\n"
			"			#11-(format2)rtr: 1 byte, value: 0-rtr Disable, 1-rtr Enable\n"
			"			#12-(format2)dlc: 1 byte, value: 0 ~ 8\n"
			"			#13-(format2)4/8 can data(0~3): 4 byte, ex: 0x01020304\n"
			"			#14-(format2)4/8 can data(4~7): 4 byte, ex: 0x05060708\n"
			"			#15-delay time after seding: 1 byte, millisecond time, ex:100 -- delay 100ms\n"
			"  -j, --j1939_test\n"
			"		0 - read data\n"
			"		1 - write/read test\n"
			"			#1-port: 1 byte\n"
			"			#2-pgn: 4 bytes, ex: 0x18FEF100\n"
			"			#3-dlc: 1 byte\n"
			"			#4-4/8 J1939 data(0~3): 4 byte, ex: 0x01020304\n"
			"			#5-4/8 J1939 data(4~7): 4 byte, ex: 0x05060708\n"
			/*
			"		1 - port config\n"
			"			#1-port: 0 or 1\n"
			"		2 - filter config\n"
			"			#1-type: 1 byte\n"
			"				0x00: Add one filter\n"
			"				0x01: Add one filter\n"
			"				0x02: Remove one filter\n"
			"				0x03: Reset all filters\n"
			"			#2-port: 1 byte\n"
			"			#3-pgn: 4 byte, ex: 0x00112233\n"
			"		3 - get filter count\n"
			"			#1-port: 0 or 1\n"
			"		4 - filter get\n"
			"			#1-port: 0 or 1\n"*/
			"  -t, --test_cnt, set counter for testing loop\n"
			"  -u, --get_uart_canspeed, get UART baudrate and can speed\n"
			"  -U, --set_uart,	set UART baudrate\n"
			"  		uart baud rate - 0: 115200, 1: 230400, 2: 460800, 3: 512000, 4: 921600\n"
			"  -S --store_config\n"
			"		1: Store system configuration, 2: Factory Reset\n"
			"  -v, --version\n"
			"  -q, --quiet_mode, set 1 to quiet mode\n"
			"  -7, --demo_j1708\n"
			"  		0 - read J1708 21 bytes pkt\n"
			"  		1 - write J1708 21 bytes pkt\n"
			"			#1-mid: 1 byte, ex: 0x00\n"
			"			#2-priority: 1 byte, ex: 0x01\n"
			"			#3-dlc: 1 byte, ex: 0x08\n"
			"			#4-4/13 data: 4 byte, ex: 0x01020304\n"
			"			#5-4/13 data: 4 byte, ex: 0x05060708\n"
			"			#6-4/13 data: 4 byte, ex: 0x090a0b0c\n"
			"			#7-4/13 data: 4 byte, ex: 0x0d0e0f10\n"
			"			#8-4/13 data: 3 byte, ex: 0x111213FF\n"
			"			    --> data will be combined as 0x01020304 0x05060708 0x090A0B0C 0x0D0E0F10 0x111213FF \n"
			"		2 - filter config\n"
			"			#1-type: 1 byte\n"
			"				0x01 : set filter\n"
			"				0xFF : reset all filter\n"
			"			#2-mid: 1 byte, ex: 0x00\n"
			"		3 - get filter count\n"
			"		4 - filter get\n"			
			"  -e, --ecompass\n"
			"  		0 - set Accelerator Full Scale\n"
			"			-scale value: 0: 2, 1: 4, 2: 8, 3: 16\n"
			"  		1 - Get Accelerator Full Scale\n"
			"  -A, --Algorithm\n"
			"  		0 - Clear sensor algorithm\n"
			"  		1 - Set Fordward/Backward Algorithm Enable/Disable\n"
			"			-switch: 0: Disable, 1: Enable\n"
			"  -k, --read_sku\n"
			"  -q, --quiet mode, set 1 to quiet mode\n"
			"\n",
			argv[0]);
	exit(0);
}

static int process_options(int argc, char *argv[])
{
	unsigned int tmp = 0;

	if(argc < 2) {
		display_help(argc, argv);
		return -1;
	}

	for (;;) {
		int option_index = 0;
		static const char *short_options = "q:hp:sd:gc:bm:fw:x:r:j:lt:vuU:S:a:7:e:A:k";
		static const struct option long_options[] = {
			{"help", no_argument, 0, 'h'},
			{"port", required_argument, 0, 'p'},
			{"scan", no_argument, 0, 's'},
			{"deinit", required_argument, 0, 'd'},
			{"get_fw_ver", no_argument, 0, 'g'},
			{"can_config", required_argument, 0, 'c'},
			{"reset_module", no_argument, 0, 'b'},
			{"mode_active", required_argument, 0, 'm'},
			{"filter_config", no_argument, 0, 'f'},
			{"filter_config_raw", required_argument, 0, 'w'},
			{"get_filter_config", required_argument, 0, 'x'},
			{"read_write_raw", required_argument, 0, 'r'},
			{"j1939_test", required_argument, 0,'j'},
			{"test_cnt", required_argument, 0, 't'},
			{"version", no_argument, 0, 'v'},
			{"quiet_mode", required_argument, 0, 'q'},
			{"set_uart", required_argument, 0, 'U'},
			{"get_uart_canspeed", no_argument, 0, 'u'},
			{"store_config",required_argument,0,'S'},
			{"set_request_time",required_argument,0,'a'},
			{"demo_j1708",required_argument,0,'7'},
			{"ecompass",required_argument,0,'e'},
			{"Algorithm",required_argument,0,'A'},
			{"read_sku",no_argument,0,'k'},
			{0,0,0,0},
		};

		int c = getopt_long(argc, argv, short_options, long_options, &option_index);
		if (c == EOF) {
			break;
		}

		switch (c) {
		case 'h':
			display_help(argc, argv);
			break;

		case 'p':
			port_name = strdup(optarg);
			break;

		case 's':
			scan_port = 1;
			break;

		case 'd':
			deinit_in_seconds = atoi(optarg);
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "deinit after %d seconds\n", deinit_in_seconds);
			break;

		case 'g':
			get_can_fw_ver = 1;
			break;

		case 'c':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				can_config = 1;
				memset(&can_config_data, 0x00, sizeof(can_config_data));

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0:
							can_config_data.port = atoi(argv[index]);
							if(can_config_data.port < 0 || can_config_data.port > 1)
								goto c_err;

							break;

						case 1:
							can_config_data.speed = atoi(argv[index]);
							if(can_config_data.speed < 0 || can_config_data.speed > 5)
								goto c_err;

							break;

						default:
							goto c_err;
					}

					index++;
					cnt++;
				};

				arg_num = FARO_CAN_SDK_CAN_CONFIG_ARGS;

				if(cnt != arg_num)
					goto c_err;

				optind = index - 1;
				goto c_ok;

c_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

c_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -c arguments done\n");
			}
			break;

		case 'b':
			reset_module = 1;
			break;

		case 'm':
		{
			int cnt = 0;
			int index = optind - 1;
			int arg_num = 0;
			while(index < argc) {
				if(argv[index][0] == '-') {
					break;
				}

				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);
				
				switch(cnt) {
					case 0:	//port
						if(atoi(argv[index]) < 0 || atoi(argv[index]) > 1)
						{
							fprintf(stdout,
							"invalid argument number.\nPort: 0 or 1\nMode: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR OBD2, 5:CAN2ADR J1939, 8:J1708_21 9:J1708_21_OFF\n");
							return -1;
						}
						else
						{
							mode_active_port = atoi(argv[index]);
						}
						break;

					case 1:	//mode
						//if(atoi(argv[index]) < 0 || atoi(argv[index]) > 10)
						if(atoi(argv[index]) != 0 && atoi(argv[index]) != 1 && atoi(argv[index]) != 2 && atoi(argv[index]) != 3 && atoi(argv[index]) != 4 && atoi(argv[index]) != 5 && atoi(argv[index]) != 8 && atoi(argv[index]) != 9)
						{
							fprintf(stdout,
							"invalid argument number.\nPort: 0 or 1\nMode: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR OBD2, 5:CAN2ADR J1939, 8:J1708_21 9:J1708_21_OFF\n");
							return -1;
						}
						else
						{
							mode_active_arg = atoi(argv[index]);
							mode_active = 1;
						}
						break;

					//default:
						
				}
				
				index++;
				cnt++;
				
				arg_num = FARO_CAN_SDK_CAN_CONFIG_ARGS;
			}
			if(cnt != arg_num)
				goto m_err;

			optind = index - 1;
			goto m_ok;

m_err:
			fprintf(stderr, "invalid argument number, should be %d\n", arg_num);

m_ok:
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "checking -m arguments done\n");
			break;
		}
			
		case 'a':
		{
			int cnt = 0;
			int index = optind - 1;
			int arg_num = 0;
			while(index < argc) {
				if(argv[index][0] == '-') {
					break;
				}

				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

				tmp = 0;

				switch(cnt) {
					case 0:	//port
						if(atoi(argv[index]) < 0 || atoi(argv[index]) > 1)
						{
							fprintf(stdout,"invalid argument number. Port: 0 or 1\n");
						}
						else
						{
							set_request_time_port = atoi(argv[index]);
						}
						break;

					case 1:	//time
						if(atoi(argv[index]) < 0 || atoi(argv[index]) > 100)
						{
							fprintf(stdout,"invalid argument number. time(ms): 0~100\n");
							return -1;
						}
						else
						{
							set_request_time_time = atoi(argv[index]);
							set_request = 1;
						}
						break;

					//default:
						
				}

				index++;
				cnt++;
			};
			arg_num = FARO_CAN_SDK_CAN_REQUEST_ARGS;

			if(cnt != arg_num)
				goto a_err;

			optind = index - 1;
			goto a_ok;

a_err:
			fprintf(stderr, "invalid argument number, should be %d\n", arg_num);

a_ok:
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "checking -a arguments done\n");
		}
		break;	

		case 'f':
			filter_config = 1;
			break;

		case 'w':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				filter_config_raw = 1;
				memset(&can_filter_config, 0x00, sizeof(can_filter_config));

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0: // 0 - ID+Mask, 1 - ID List, 2 - Remove, 3 - Reset All
							can_filter_config.type = atoi(argv[index]);
							if(can_filter_config.type < 0 || can_filter_config.type > 3)
								goto w_err;

							break;

						case 1:
							can_filter_config.port = atoi(argv[index]);
							if(can_filter_config.port < 0 && can_filter_config.port > 1)
								goto w_err;

							break;

						case 2:
							can_filter_config.bank = atoi(argv[index]);
							if(can_filter_config.bank < 0 || can_filter_config.bank > 13)
								goto w_err;

							break;

						case 3:
							can_filter_config.mode = atoi(argv[index]);
							if(can_filter_config.mode < 0 || can_filter_config.mode > 1)
								goto w_err;

							break;

						case 4:
							if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
								sscanf(argv[index], "0x%08X", &can_filter_config.filterId);
							} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterId |= (tmp << 16);
							}
							break;

						case 5:
							if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
								sscanf(argv[index], "0x%08X", &can_filter_config.filterMask);
							} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterId |= tmp;
							}
							break;

						case 6:
							if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterMask |= (tmp << 16);
							}
							break;

						case 7:
							if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterMask |= tmp;
							}
							break;

						default:
							goto w_err;
					}

					index++;
					cnt++;
				};

				if((can_filter_config.type == 2) || (can_filter_config.type == 3)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RESET_REMOVE;
				} else if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
					(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_32_BITS;
				} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
					(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_16_BITS;
				}

				if(cnt != arg_num)
					goto w_err;

				optind = index - 1;
				goto w_ok;

w_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

w_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -w arguments done\n");
			}
			break;

		case 'x':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				get_filter_config = 1;
				memset(&can_filter_config, 0x00, sizeof(can_filter_config));

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					switch(cnt) {
						case 0: // port
							can_filter_config.port = atoi(argv[index]);
							if(can_filter_config.port < 0 || can_filter_config.port > 1)
							{
								fprintf(stdout,"invalid argument number.\nPort value: 0 or 1\n");
								return -1;
							}

							break;

						case 1: // bank
							can_filter_config.bank = atoi(argv[index]);
							if(can_filter_config.bank < 0 || can_filter_config.bank > 13)
							{
								fprintf(stdout,"invalid argument number.\nBank value: 0 ~ 13\n");
								return -1;
							}

							break;
					}

					index++;
					cnt++;
				};

				arg_num = FARO_CAN_SDK_GET_FILTER_CONFIG;
				if(cnt != arg_num)
					goto x_err;

				optind = index - 1;
				goto x_ok;

x_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

x_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -x arguments done\n");
			}
			break;

		case 'r':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;
				//read_write_raw = 1;
				rawcan_demo = 1;

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0: // 0: write/read, 1: port config, 2: filter config, 3: get port count
							rawcan_demo_arg = atoi(argv[index]);
							if(rawcan_demo_arg < 0 || rawcan_demo_arg > NUM_OF_DEMO_RAWCAN_TEST_IDX)
							{
								fprintf(stdout,"invalid argument number.\nValue: 0 ~ %d\n",NUM_OF_DEMO_RAWCAN_TEST_IDX);
								return -1;
							}

							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								memset(&rawcan_send_msg[0], 0x00, sizeof(rawcan_send_msg[0]));
								memset(&rawcan_send_msg[1], 0x00, sizeof(rawcan_send_msg[1]));
							}
							break;

						case 1:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg ==DEMO_TEST_IDX_RAWCAN_WRITE_READ
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "%d", &tmp);
								rawcan_send_msg[0].port = tmp;			
							}
							break;

						case 2:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "0x%08X", &tmp);
								//memcpy(&rawcan_send_msg.id, &tmp, sizeof(uint32_t));
								rawcan_send_msg[0].id = tmp;
							} 
							break;

						case 3:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								//sscanf(argv[index], "0x%02X", &tmp);
								sscanf(argv[index], "%d", &tmp);
								rawcan_send_msg[0].mode = tmp;
							}
							break;

						case 4:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "%d", &tmp);
								rawcan_send_msg[0].rtr = tmp;
							}
							break;

						case 5:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "%d", &tmp);
								rawcan_send_msg[0].dlc = tmp;
							}
							break;

						case 6:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&rawcan_send_msg[0].data[0], &tmp, sizeof(uint32_t));
							}
							break;

						case 7:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&rawcan_send_msg[0].data[4], &tmp, sizeof(uint32_t));
							}
							break;

						case 8:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ) {
								sscanf(argv[index], "%d", &tmp);
								rawcan_delay_time = tmp * 1000;	//change us to ms
							}
							else if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "%d", &tmp);
								rawcan_send_msg[1].port = tmp;
							}
							break;	

						case 9:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "0x%08X", &tmp);
								//memcpy(&rawcan_send_msg.id, &tmp, sizeof(uint32_t));
								rawcan_send_msg[1].id = tmp;
							} 
							break;

						case 10:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								//sscanf(argv[index], "0x%02X", &tmp);
								sscanf(argv[index], "%d", &tmp);
								rawcan_send_msg[1].mode = tmp;
							}
							break;

						case 11:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "%d", &tmp);
								rawcan_send_msg[1].rtr = tmp;
							}
							break;

						case 12:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "%d", &tmp);
								rawcan_send_msg[1].dlc = tmp;
							}
							break;

						case 13:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&rawcan_send_msg[1].data[0], &tmp, sizeof(uint32_t));
							}
							break;

						case 14:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&rawcan_send_msg[1].data[4], &tmp, sizeof(uint32_t));
							}
							break;

						case 15:
							if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE
								|| rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ) {
								sscanf(argv[index], "%d", &tmp);
								rawcan_delay_time = tmp * 1000;	//change us to ms
							}
							break;	
							
						default:
							goto rawcan_err;
					}
 
					index++;
					cnt++;
				};
				
					//fprintf(stdout,"rawcan_demo_arg = %d",rawcan_demo_arg);				
					
				if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_READ) {
					arg_num = FARO_CAN_SDK_RAWCAN_READ_ARGS;
				}else if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE) {
					arg_num = FARO_CAN_SDK_RAWCAN_WRITE_ARGS;
				}else if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE) {
					arg_num = FARO_CAN_SDK_RAWCAN_2PORT_WRITE_ARGS;
				}else if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_WRITE_READ){
					arg_num = FARO_CAN_SDK_RAWCAN_WRITE_READ_ARGS;
				}else if(rawcan_demo_arg == DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ){
					arg_num = FARO_CAN_SDK_RAWCAN_2PORT_WRITE_READ_ARGS;
				}
				//fprintf(stdout, "cnt = %d, arg_num = %d\n",cnt,arg_num);
				if(cnt != arg_num)
					goto rawcan_err;

				optind = index - 1;
				goto rawcan_ok;

rawcan_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				rawcan_demo = 0;
				//display_help(argc, argv);

rawcan_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -j arguments done\n");
			}
			break;			

		case 'j':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				j1939_test = 1;

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {// 0: Read, 1: Write and read
						case 0: 
							j1939_test_arg = atoi(argv[index]);
							if(j1939_test_arg < 0 || j1939_test_arg > NUM_OF_DEMO_J1939_TEST_IDX)
							{
								fprintf(stdout,"invalid argument number.\nValue: 0 ~ %d\n",NUM_OF_DEMO_J1939_TEST_IDX);
								return -1;
							}

							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								memset(&j1939_send_msg, 0x00, sizeof(j1939_send_msg));
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
								memset(&j1939_filter_config, 0x00, sizeof(j1939_filter_config));
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_GET) {
								j1939_test_port = 0;
							}else if(j1939_test_arg == J1939_TEST_IDX_READ) {
								j1939_test_port = 0;
							}

							break;

						case 1:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.port = tmp;
							} else if(j1939_test_arg == J1939_TEST_IDX_PORT_CONFIG) {
								j1939_test_port = atoi(argv[index]);
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_filter_config.type = tmp;
							} else if(j1939_test_arg == J1939_TEST_IDX_GET_FILTER_COUNTER) {
								j1939_test_port = atoi(argv[index]);
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_GET) {
								j1939_test_port = atoi(argv[index]);
							}
							break;

						case 2:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								//sscanf(argv[index], "0x%02X", &tmp);
								//j1939_send_msg.p = tmp;
								sscanf(argv[index], "0x%08X", &tmp);
								memcpy(&j1939_send_msg.pgn, &tmp, sizeof(uint32_t));
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_filter_config.port = tmp;
							}
							break;

						case 3:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								//sscanf(argv[index], "0x%02X", &tmp);
								//j1939_send_msg.sa = tmp;
								sscanf(argv[index], "%d", &tmp);
								j1939_send_msg.dlc = tmp;
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
								sscanf(argv[index], "0x%08X", &tmp);
								//j1939_filter_config.pgn = tmp;
								//tmp = ntohl(tmp);
								memcpy(&j1939_filter_config.pgn, &tmp, sizeof(uint32_t));
							}
							break;

						case 4:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								//sscanf(argv[index], "0x%08X", &tmp);
								//memcpy(&j1939_send_msg.pgn, &tmp, sizeof(uint32_t));
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&j1939_send_msg.data[0], &tmp, sizeof(uint32_t));
							}
							break;

						case 5:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								//sscanf(argv[index], "0x%02X", &tmp);
								//j1939_send_msg.dlc = tmp;
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&j1939_send_msg.data[4], &tmp, sizeof(uint32_t));
							}
							break;

						case 6:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%08X", &tmp);
								j1939_send_msg.data[0] = tmp;
							}
							break;

						case 7:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[1] = tmp;
							}
							break;

						case 8:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[2] = tmp;
							}
							break;

						case 9:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[3] = tmp;
							}
							break;

						case 10:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[4] = tmp;
							}
							break;

						case 11:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[5] = tmp;
							}
							break;

						case 12:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[6] = tmp;
							}
							break;

						case 13:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[7] = tmp;
							}
							break;

						default:
							goto j_err;
					}

					index++;
					cnt++;
				};

				if(j1939_test_arg == J1939_TEST_IDX_RW) {
					arg_num = FARO_CAN_SDK_J1939_RW_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_PORT_CONFIG) {
					arg_num = FARO_CAN_SDK_J1939_PORT_CONFIG_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
					arg_num = FARO_CAN_SDK_J1939_FILTER_CONFIG_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_GET_FILTER_COUNTER) {
					arg_num = FARO_CAN_SDK_J1939_GET_FILTER_COUNTER_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_GET) {
					arg_num = FARO_CAN_SDK_J1939_FILTER_GET_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_READ) {
					arg_num = FARO_CAN_SDK_J1939_READ;
				}

				if(cnt != arg_num)
					goto j_err;

				optind = index - 1;
				goto j_ok;

j_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

j_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -j arguments done\n");
			}
			break;

		case 't':
			test_cnt = atoi(optarg);
			if(test_cnt <= 0)
				test_cnt = 1;
			break;

		case 'u':
			get_uart_canbaudrate = 1;		
			break;
			
		case 'U':
			if(atoi(optarg) > 4 || atoi(optarg) < 0)
			{
				fprintf(stdout, "Set uart baudrate error.\n 0: 115200, 1: 230400, 2: 460800, 3: 512000, 4: 921600.\n");
				return -1;
			}
			else
			{
				set_uart_baudrate = 1;
				set_uart_baudrate_arg = atoi(optarg);
				sleep(1);
			}
			break;
			
		case 'S':
			if(atoi(optarg) > 3 || atoi(optarg) < 1)
			{
				fprintf(stdout, "invalid argument number.\n 1: Store system configuration, 2: Factory Reset\n.");
				return -1;
			}
			else
			{
				store_config = 1;
				store_config_mode = atoi(optarg);
				sleep(1);
			}
			break;
		
		case 'v':
			fprintf(stdout, "%s Version %d.%d.%d\n",
				argv[0],
				FARO_CAN_SDK_DEMO_VERSION_MAJOR,
				FARO_CAN_SDK_DEMO_VERSION_MINOR,
				FARO_CAN_SDK_DEMO_VERSION_PATCH);
			return -1;

		case 'q':
			g_quiet_print_flag = atoi(optarg);
			sleep(1);
			if(g_quiet_print_flag > 1 || g_quiet_print_flag < 0)
			{
				fprintf(stdout, "invalid argument number.\n0:Normal mode, 1:Quiet mode\n");
				return -1;
			}
			break;

		case '7':
		{
			int cnt = 0;
			int index = optind - 1;
			int arg_num = 0;

			j1708_21_demo = 1;

			while(index < argc) {
				if(argv[index][0] == '-') {
					break;
				}

				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

				tmp = 0;

				switch(cnt) {
					case 0:
						j1708_21_demo_arg = atoi(argv[index]);
						if(j1708_21_demo_arg < 0 || j1708_21_demo_arg >= NUM_OF_DEMO_J1708_TEST_IDX)
						{
							fprintf(stdout,"invalid argument number.\nValue: 0 ~ 5\n");
							return -1;
						}
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							memset(&j1708_21_send_msg, 0x00, sizeof(j1708_21_send_msg));
						} else if(j1708_21_demo_arg == J1708_TEST_IDX_FILTER_CONFIG) {
							memset(&j1708_filter_config, 0x00, sizeof(j1708_filter_config));							
						}
						break;

					case 1:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							sscanf(argv[index], "0x%02X", &tmp);
							j1708_21_send_msg.mid = tmp;
						} else if(j1708_21_demo_arg == J1708_TEST_IDX_FILTER_CONFIG) {								
							sscanf(argv[index], "0x%02X", &tmp);								
							j1708_filter_config.type = tmp;							
						}
						break;

					case 2:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							sscanf(argv[index], "0x%02X", &tmp);
							j1708_21_send_msg.p = tmp;
						} else if(j1708_21_demo_arg == J1708_TEST_IDX_FILTER_CONFIG) {								
							sscanf(argv[index], "0x%02X", &tmp);								
							j1708_filter_config.mid = tmp;							
						}
						break;

					case 3:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							sscanf(argv[index], "0x%02X", &tmp);
							j1708_21_send_msg.dlc = tmp;
						}
						break;

					case 4:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							sscanf(argv[index], "0x%08X", &tmp);
							tmp = ntohl(tmp);
							//memcpy(&j1708_21_send_msg.data[0], &tmp, sizeof(uint32_t));
							memcpy(&j1708_send_allmsg.alldata[0], &tmp, sizeof(uint32_t));
						}
						break;

					case 5:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							sscanf(argv[index], "0x%08X", &tmp);
							tmp = ntohl(tmp);
							//memcpy(&j1708_21_send_msg.data[4], &tmp, sizeof(uint32_t));
							memcpy(&j1708_send_allmsg.alldata[4], &tmp, sizeof(uint32_t));
						}
						break;

					case 6:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							sscanf(argv[index], "0x%08X", &tmp);
							tmp = ntohl(tmp);
							//memcpy(&j1708_21_send_msg.data[8], &tmp, sizeof(uint32_t));
							memcpy(&j1708_send_allmsg.alldata[8], &tmp, sizeof(uint32_t));
						}
						break;

					case 7:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							sscanf(argv[index], "0x%08X", &tmp);
							tmp = ntohl(tmp);
							//memcpy(&j1708_21_send_msg.data[12], &tmp, (sizeof(uint32_t) - 3));
							memcpy(&j1708_send_allmsg.alldata[12], &tmp, sizeof(uint32_t));
						}
						break;

					case 8:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							sscanf(argv[index], "0x%08X", &tmp);
							tmp = ntohl(tmp);
							//memcpy(&j1708_21_send_msg.data[12], &tmp, (sizeof(uint32_t) - 3));
							memcpy(&j1708_send_allmsg.alldata[16], &tmp, sizeof(uint32_t));
						}
						break;
					/*
					case 1:
						if(j1708_21_demo_arg == DEMO_TEST_IDX_J1708_WRITE_21) {
							j1708_21_send_msg.pkt_no = atoi(argv[index]);
							if((j1708_21_send_msg.pkt_no != 0) &&
								j1708_21_send_msg.pkt_no != 1)
								goto j1708_21_err;
						}
						break;
					*/
					default:
						goto j1708_21_err;
				}

				index++;
				cnt++;
			};

			switch(j1708_21_demo_arg) {
				case DEMO_TEST_IDX_J1708_READ_21:		arg_num = FARO_CAN_SDK_DEMO_J1708_READ_21;			break;
				case DEMO_TEST_IDX_J1708_WRITE_21:		arg_num = FARO_CAN_SDK_DEMO_J1708_WRITE_21;			break;
				case J1708_TEST_IDX_FILTER_CONFIG:		arg_num = FARO_CAN_SDK_J1708_FILTER_CONFIG_ARGS;		break;					
				case J1708_TEST_IDX_GET_FILTER_COUNTER:	arg_num = FARO_CAN_SDK_J1708_GET_FILTER_COUNTS_ARGS;	break;					
				case J1708_TEST_IDX_FILTER_GET:			arg_num = FARO_CAN_SDK_J1708_FILTER_GET_ARGS;			break;
			}

			//check is all argv is correct.
			if(cnt != arg_num)
			{
				j1708_21_demo = 0;
				goto j1708_21_err;
			}

			optind = index - 1;
			goto j1708_21_ok;

j1708_21_err:
			fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
			return -1;

j1708_21_ok:
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "checking -t arguments done\n");
		}
		break;

		case 'e':
		{
			int cnt = 0;
			int index = optind - 1;
			int arg_num = 0;

			while(index < argc) {
				if(argv[index][0] == '-') {
					break;
				}

				tmp = 0;

				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

				switch(cnt) {
					case 0:
						ecompass_demo_arg = atoi(argv[index]);
						if(ecompass_demo_arg < 0 || ecompass_demo_arg >= NUM_OF_DEMO_ecompass_IDX)
						{
							fprintf(stderr, "invalid argument number, should be 0 or 1\n");
							return -1;
						}
						else if(ecompass_demo_arg == DEMO_TEST_IDX_ecompass_get_accelerator)
						{
							ecompass_demo = 1;
						}
						break;

					case 1:
						if(ecompass_demo_arg == DEMO_TEST_IDX_ecompass_set_accelerator) {
							ecompass_acc_scale = atoi(argv[index]);
							if(ecompass_acc_scale < 0 || ecompass_acc_scale > 3)
							{
								fprintf(stderr, "invalid argument number, should be 0 ~ 3\n");
								return -1;
							}
							ecompass_demo = 1;
						}
						break;

					default:
						goto ecompass_err;
				}

				index++;
				cnt++;
			};
			
			switch(ecompass_demo_arg) {
				case DEMO_TEST_IDX_ecompass_set_accelerator:		arg_num = FARO_CAN_SDK_DEMO_ecompass_set_acu;			break;
				case DEMO_TEST_IDX_ecompass_get_accelerator:		arg_num = FARO_CAN_SDK_DEMO_ecompass_get_acu;			break;
			}

			//check is all argv is correct.
			if(cnt != arg_num)
			{
				goto ecompass_err;
			}

			optind = index - 1;
			goto ecompass_ok;

ecompass_err:
			fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
			return -1;

ecompass_ok:
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "checking -t arguments done\n");
		}
		break;

		case 'A':
		{
			int cnt = 0;
			int index = optind - 1;
			int arg_num = 0;

			while(index < argc) {
				if(argv[index][0] == '-') {
					break;
				}

				tmp = 0;

				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

				switch(cnt) {
					case 0:
						algorithm_demo_arg = atoi(argv[index]);
						if(algorithm_demo_arg < 0 || algorithm_demo_arg >= NUM_OF_DEMO_ALGORITHM_IDX)
						{
							fprintf(stderr, "invalid argument number, should be 0 ~ 2\n");
							return -1;
						}
						else if(algorithm_demo_arg == DEMO_TEST_IDX_ALGORITHM_CLEAR_G0Gf
							||	algorithm_demo_arg == DEMO_TEST_IDX_ALGORITHM_GET_F_B_SWH)
						{
							algorithm_demo = 1;
						}
						break;

					case 1:
						if(algorithm_demo_arg == DEMO_TEST_IDX_ALGORITHM_SET_F_B_SWH) {
							algorithm_enable = atoi(argv[index]);
							if(algorithm_enable < 0 || algorithm_enable > 1)
							{
								fprintf(stderr, "invalid argument number, 0 : Disable, 1 : Enable\n");
								return -1;
							}
							algorithm_demo = 1;
						}
						break;

					default:
						goto algorithm_err;
				}

				index++;
				cnt++;
			};
			
			switch(algorithm_demo_arg) {
				case DEMO_TEST_IDX_ALGORITHM_CLEAR_G0Gf:		arg_num = FARO_CAN_SDK_DEMO_algorithm_clear_gogf;			break;
				case DEMO_TEST_IDX_ALGORITHM_SET_F_B_SWH:		arg_num = FARO_CAN_SDK_DEMO_algorithm_set_fb_swh;			break;
				case DEMO_TEST_IDX_ALGORITHM_GET_F_B_SWH:		arg_num = FARO_CAN_SDK_DEMO_algorithm_get_fb_swh;			break;
			}
			
			//check is all argv is correct.
			if(cnt != arg_num)
			{
				goto algorithm_err;
			}

			optind = index - 1;
			goto algorithm_ok;

algorithm_err:
			fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
			return -1;

algorithm_ok:
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "checking -t arguments done\n");
		}
		break;

		case 'k':
			read_sku = 1;
		break;

		}
	}

	return 0;
}

static void *deinit_wait(void *args)
{
	int i;

	if(deinit_in_seconds == -1) {
		while(stop_thread == 0)
			sleep(1);
	} else {
		for(i=0; i<deinit_in_seconds; i++) {
			if(stop_thread != 0)
				break;

			sleep(1);
		}
	}

	if(stop_thread == 0) {
		stop_thread = 1;
		sleep(1);
	}

	AZ_VC_DeInit();

	pthread_exit(NULL);
}

void sig_handler(int signo)
{
	if(signo == SIGINT) {
		fprintf(stderr, "catch SIGINT\n");
		stop_thread = 1;
	}

	if(launch_receiver) {
		AZ_VC_CAN_Dump_Pkt_Cnt();
	}
}

static int do_scan_port(void)
{
	int ret = 0;
	char buf[PATH_MAX];
	int serial_port_num = 0;

	memset(buf, sizeof(buf), 0x00);
	AZ_VC_GetCommPort(buf, &serial_port_num);
	fprintf(stdout, "Found %d serial ports\n", serial_port_num);
	if(serial_port_num > 0)
		fprintf(stdout, "buf -> \n%s", buf);

	return ret;
}

static void *do_get_can_fw_ver(void *args)
{
	int ret = 0;
	int i;
	struct can_fw_ver_t can_fw_ver;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting get firmware version on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail, test_cnt = %d\n", i);
			continue;
		}

		fprintf(stdout, "get FW count = %d, can_fw_ver: %02X.%02X.%02X.%02X.%02X\n",
				i+1, can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
	}

err:
	pthread_exit(NULL);
}

static void *do_get_uart_canbaudrate(void *args)
{
	int ret = 0;
	int i;
	struct uart_canbaudrate_t uart_canbaudrate;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting get get_uart_canbaudrate on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	memset(&uart_canbaudrate, 0x00, sizeof(uart_canbaudrate));


	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;
		if((ret = AZ_VC_GetUART_CanBaudRate(&uart_canbaudrate)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetUART_CanBaudRate fail\n");
			ret = -1;
			goto err;
		}

		fprintf(stdout,"UART BaudRate = ");
		switch(uart_canbaudrate.UART){
			case 0:
				fprintf(stdout,"115200");
				break;
			case 1:
				fprintf(stdout,"230400");
				break;
			case 2:
				fprintf(stdout,"460800");
				break;
			case 3:
				fprintf(stdout,"512000");
				break;
			case 4:
				fprintf(stdout,"921600");
				break;
		}
		fprintf(stdout,".\n");
		fprintf(stdout,"Port0 Can Speed = ");
		switch(uart_canbaudrate.CANPort0){
			case 0:
				fprintf(stdout,"1M Bits/s");
				break;
			case 1:
				fprintf(stdout,"800k Bits/s");
				break;
			case 2:
				fprintf(stdout,"500k Bits/s");
				break;
			case 3:
				fprintf(stdout,"250k Bits/s");
				break;
			case 4:
				fprintf(stdout,"200k Bits/s");
				break;
			case 5:
				fprintf(stdout,"125k Bits/s");
				break;
			default:
				fprintf(stdout,"NULL");
				break;
		}
		fprintf(stdout,".\n");
		fprintf(stdout,"Port1 Can Speed = ");
		switch(uart_canbaudrate.CANPort1){
			case 0:
				fprintf(stdout,"1M Bits/s");
				break;
			case 1:
				fprintf(stdout,"800k Bits/s");
				break;
			case 2:
				fprintf(stdout,"500k Bits/s");
				break;
			case 3:
				fprintf(stdout,"250k Bits/s");
				break;
			case 4:
				fprintf(stdout,"200k Bits/s");
				break;
			case 5:
				fprintf(stdout,"125k Bits/s");
				break;
			default:
				fprintf(stdout,"NULL");
				break;
		}
		fprintf(stdout,".\n");
		
		fprintf(stdout,"Port 0 Can Mode = %d", uart_canbaudrate.CANPort0_Mode);
		fprintf(stdout,",\n");
		fprintf(stdout,"Port 1 Can Mode = %d", uart_canbaudrate.CANPort1_Mode);
		fprintf(stdout,",\n");
		fprintf(stdout,"J1708 Mode = %d", uart_canbaudrate.J1708_Mode);
		fprintf(stdout,",\n");
		fprintf(stdout,"J1708 retry cnt count = %d", uart_canbaudrate.J1708_retry_count);
		fprintf(stdout,",\n");
		fprintf(stdout,"request time = %d", uart_canbaudrate.request_time);
		
		fprintf(stdout,".\n");
		
		stop_thread = 1;
	}

err:
	pthread_exit(NULL);
}

static void *do_set_uart_baudrate(void *args)
{
	int ret = 0;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_set_uart_baudrate on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;
		
	ret = AZ_VC_UART_Config(set_uart_baudrate_arg);
	if(ret != 0)
	{
		fprintf(stderr, "Calling AZ_VC_UART_Config fail\n");
	}
	else
	{
		fprintf(stdout, "Calling AZ_VC_UART_Config success\n");
	}
	stop_thread = 1;
err:
	pthread_exit(NULL);
}

static void *do_reset_module(void *args)
{
	int ret = 0;
	int i;

	deinit_in_seconds = 90;
	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;
	
	struct can_filter_config_t can_filter_config;
	memset(&can_filter_config, 0x00, sizeof(can_filter_config));
	
	can_filter_config.type = 0x03;
	can_filter_config.port = 0x00;
	can_filter_config.bank = 0x07;
	can_filter_config.mode = 0x00;
	AZ_VC_CAN_Filter_Config(can_filter_config);
	
	struct can_fw_ver_t can_fw_ver;

	memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
	
	//do reset module 3 times
	for(i = 0; i < 3; i++)
	{	
		if(stop_thread != 0)
			break;
		AZ_VC_Reset_Module();
		sleep(5);
	}
	
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling reset module fail.");
			continue;
		} 
		
		fprintf(stdout, "reset module finish.\n");
	}	
	stop_thread = 1;
err:
	pthread_exit(NULL);
}

static void *do_can_config(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting can_config on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X.%02X.%02X\n", 
			can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		if((ret = AZ_VC_CAN_Config(can_config_data)) != 0)
			goto err;

		fprintf(stdout, "test_cnt = %d, port = %d, speed = %d\n",
			i, can_config_data.port, can_config_data.speed);
	}

err:
	stop_thread = 1;
	pthread_exit(NULL);
}

static void *do_mode_active(void *args)
{
	int ret = 0;
	int i;
	static int current_active_mode = 0;

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		if(port_name == NULL) {
			fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
			goto err;
		}
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;
		
		if(mode_active_arg == 4 || mode_active_arg == 5)
		{
			fprintf(stdout, "Active ADR and checking...\n");
		}
		usleep(1);
		if((ret = AZ_VC_ModeActive(mode_active_port, mode_active_arg)) != 0) {
			fprintf(stderr, "calling AZ_VC_ModeActive fail\n");
			goto err;
		}

		fprintf(stdout, "mode_active argument : %d success.\n", mode_active_arg);
	}
	stop_thread = 1;

err:
	pthread_exit(NULL);
}

static void *do_filter_config(void *args)
{
	int ret = 0;
	int i, j;
	struct can_filter_config_t can_filter_config[] = {
		/*{2, 0, 0, 0, 0, 0},
		{2, 1, 0, 0, 0, 0},
		{2, 2, 0, 0, 0, 0},
		{2, 3, 0, 0, 0, 0},
		{2, 4, 0, 0, 0, 0},
		{2, 5, 0, 0, 0, 0},
		{2, 6, 0, 0, 0, 0},
		{2, 7, 0, 0, 0, 0},
		{2, 8, 0, 0, 0, 0},
		{2, 9, 0, 0, 0, 0},
		{2, 10, 0, 0, 0, 0},
		{2, 11, 0, 0, 0, 3},
		{2, 12, 0, 0, 0, 2},
		{2, 13, 0, 0, 0, 1},*/
		{2, 0, 0, 0, 0, 0},
		{2, 0, 1, 0, 0, 0},
		{2, 0, 2, 0, 0, 0},
		{2, 0, 3, 0, 0, 0},
		{2, 0, 4, 0, 0, 0},
		{2, 0, 5, 0, 0, 0},
		{2, 0, 6, 0, 0, 0},
		{2, 0, 7, 0, 0, 0},
		{2, 0, 8, 0, 0, 0},
		{2, 0, 9, 0, 0, 0},
		{2, 0, 10, 0, 0, 0},
		{2, 0, 11, 0, 0, 3},
		{2, 0, 12, 0, 0, 2},
		{2, 0, 13, 0, 0, 1},
	};

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_filter_config on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X.%02X.%02X\n", 
			can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
	}


	for(i=0; i<test_cnt; i++) {
		for(j=0; j<(sizeof(can_filter_config)/sizeof(struct can_filter_config_t)); j++) {
			if(stop_thread != 0)
				break;

			if((ret = AZ_VC_CAN_Filter_Config(can_filter_config[j])) != 0)
				goto err;

			fprintf(stdout, "test_cnt = %d, type = %d, port = %d, bank = %d, filterId = %d, filterMask = %d\n",
				i, can_filter_config[j].type, can_filter_config[j].port, can_filter_config[j].bank, can_filter_config[j].filterId, can_filter_config[j].filterMask);
		}
	}
	fprintf(stdout, "Set filter config success.\n");
		stop_thread = 1;

err:
	pthread_exit(NULL);
}

static void *do_filter_config_raw(void *args)
{
	int ret = 0;
	struct can_filter_config_t *filter = (struct can_filter_config_t *)args;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "type = %d, port = %d, bank = %d, mode = %d, filterId = 0x%08X, filerMask = 0x%08X\n",
			filter->type, filter->port, filter->bank, filter->mode, filter->filterId, filter->filterMask);

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_filter_config_raw on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

/*
	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}
*/

	if((ret = AZ_VC_CAN_Filter_Config(*filter)) != 0)
		goto err;
	else
	{
		
		fprintf(stdout, "Set filter config success.\n");
		stop_thread = 1;
	}

err:
	pthread_exit(NULL);
}

static void *do_get_filter_config(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_get_filter_config on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X.%02X.%02X\n", 
			can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		fprintf(stdout, "set port = %d, bank = %d\n",
			can_filter_config.port, can_filter_config.bank);

		if((ret = AZ_VC_CAN_Filter_Get(&can_filter_config)) != 0)
			goto err;

		fprintf(stdout, "test_cnt = %d:\n", i);
		fprintf(stdout, "\ttype = %d, port = %d, bank = %d, mode = %d\n",
			can_filter_config.type, can_filter_config.port, can_filter_config.bank, can_filter_config.mode);
		fprintf(stdout, "\tfilterId = 0x%08X, filterMask = 0x%08X\n",
			can_filter_config.filterId, can_filter_config.filterMask);
	}

err:
	pthread_exit(NULL);
}

static void *do_rawcan_demo(void *args)
{
	int ret = 0;
	int i;
	int can_index;
	int can_amount;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_rawcan_demo on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X.%02X.%02X\n", 
			can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
	}

	switch(rawcan_demo_arg) {
		case DEMO_TEST_IDX_RAWCAN_READ:
			{
				int receive_count = 0;
				while(1)	{		
					if(stop_thread != 0)			
						break;		
					struct can_message_t msg;		
					memset(&msg, 0x00, sizeof(msg));		
					uint32_t size = 0;
					can_amount = AZ_VC_CAN_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						if((ret = AZ_VC_CAN_Read(&msg, &size)) == 0) {					
							fprintf(stdout, "Receive count = %d, port = %d, ID = 0x%08X, data =", receive_count + 1, msg.port, msg.id);									
							for(i=0; i<sizeof(msg.data); i++)
								fprintf(stdout, " 0x%02X", msg.data[i]);
							fprintf(stdout,"\n");
							receive_count++;
							continue;
						}
					}
				}
			}
			break;

		case DEMO_TEST_IDX_RAWCAN_WRITE:
			{
				int k = 0;
				for(i=0; i<test_cnt; i++) {
					if(stop_thread != 0)
						break;
					
					ret = AZ_VC_CAN_Write(rawcan_send_msg[0]);
					if(ret != 0)
						goto err;
					
					fprintf(stdout, "Send count = %d, port = %d, id = 0x%08X, data = ",	i+1, rawcan_send_msg[0].port, rawcan_send_msg[0].id);

					for(k=0; k<rawcan_send_msg[0].dlc; k++)
						fprintf(stdout, " 0x%02X", rawcan_send_msg[0].data[k]);
					
					fprintf(stdout,"\n");
					usleep(rawcan_delay_time);
				}
				stop_thread = 1;
			}
			break;

		case DEMO_TEST_IDX_RAWCAN_2PORT_WRITE:
			{
				int k = 0;
				for(i=0; i<test_cnt; i++) {
					if(stop_thread != 0)
						break;
					
					ret = AZ_VC_CAN_Write(rawcan_send_msg[0]);
					if(ret != 0)
						goto err;
					
					fprintf(stdout, "Send count = %d, port = %d, id = 0x%08X, data = ",	i+1, rawcan_send_msg[0].port, rawcan_send_msg[0].id);

					for(k=0; k<rawcan_send_msg[0].dlc; k++)
						fprintf(stdout, " 0x%02X", rawcan_send_msg[0].data[k]);
					
					fprintf(stdout,"\n");
					usleep(rawcan_delay_time);

					ret = AZ_VC_CAN_Write(rawcan_send_msg[1]);
					if(ret != 0)
						goto err;
					
					fprintf(stdout, "Send count = %d, port = %d, id = 0x%08X, data = ",	i+1, rawcan_send_msg[1].port, rawcan_send_msg[1].id);

					for(k=0; k<rawcan_send_msg[1].dlc; k++)
						fprintf(stdout, " 0x%02X", rawcan_send_msg[1].data[k]);
					
					fprintf(stdout,"\n");
					usleep(rawcan_delay_time);
				}
				stop_thread = 1;
			}
			break;

		case DEMO_TEST_IDX_RAWCAN_WRITE_READ:
			{
				int k = 0, j = 0;
				int receive_count = 0;
				struct can_message_t msg;	
				for(i=0; i<test_cnt; i++) {
					if(stop_thread != 0)
						break;
					
					ret = AZ_VC_CAN_Write(rawcan_send_msg[0]);
					if(ret != 0)
						goto err;
					
					fprintf(stdout, "Send count = %d, port = %d, id = 0x%08X, data = ",	i+1, rawcan_send_msg[0].port, rawcan_send_msg[0].id);

					for(k=0; k<rawcan_send_msg[0].dlc; k++)
						fprintf(stdout, " 0x%02X", rawcan_send_msg[0].data[k]);
					
					fprintf(stdout,"\n");
					usleep(rawcan_delay_time);
					
					memset(&msg, 0x00, sizeof(msg));		
					uint32_t size = 0;		
					
					can_amount = AZ_VC_CAN_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						if((ret = AZ_VC_CAN_Read_wait(&msg, &size,10)) == 0) {
							fprintf(stdout, "Receive count = %d, port = %d, ID = 0x%08X, data =", receive_count + 1, msg.port, msg.id);									
							for(j=0; j<sizeof(msg.data); j++)
								fprintf(stdout, " 0x%02X", msg.data[j]);
							fprintf(stdout,"\n");
							receive_count++;
							continue;
						}
					}
					
				}
				stop_thread = 1;
			}
			break;			

		case DEMO_TEST_IDX_RAWCAN_2PORT_WRITE_READ:
			{
				int k = 0, j = 0;
				int receive_count = 0;
				struct can_message_t msg;	
				for(i=0; i<test_cnt; i++) {
					if(stop_thread != 0)
						break;
					
					ret = AZ_VC_CAN_Write(rawcan_send_msg[0]);
					if(ret != 0)
						goto err;
					
					fprintf(stdout, "Send count = %d, port = %d, id = 0x%08X, data = ",	i+1, rawcan_send_msg[0].port, rawcan_send_msg[0].id);

					for(k=0; k<rawcan_send_msg[0].dlc; k++)
						fprintf(stdout, " 0x%02X", rawcan_send_msg[0].data[k]);
					
					fprintf(stdout,"\n");
					usleep(rawcan_delay_time);

					ret = AZ_VC_CAN_Write(rawcan_send_msg[1]);
					if(ret != 0)
						goto err;
					
					fprintf(stdout, "Send count = %d, port = %d, id = 0x%08X, data = ",	i+1, rawcan_send_msg[1].port, rawcan_send_msg[1].id);

					for(k=0; k<rawcan_send_msg[1].dlc; k++)
						fprintf(stdout, " 0x%02X", rawcan_send_msg[1].data[k]);
					
					fprintf(stdout,"\n");
					usleep(rawcan_delay_time);
					
					memset(&msg, 0x00, sizeof(msg));		
					uint32_t size = 0;		

					can_amount = AZ_VC_CAN_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						if((ret = AZ_VC_CAN_Read_wait(&msg, &size,10)) == 0) {
							fprintf(stdout, "Receive count = %d, port = %d, ID = 0x%08X, data =", receive_count + 1, msg.port, msg.id);									
							for(j=0; j<sizeof(msg.data); j++)
								fprintf(stdout, " 0x%02X", msg.data[i]);
							fprintf(stdout,"\n");
							receive_count++;
							continue;
						}
					}

					
				}
				stop_thread = 1;
			}
			break;
			
		default:
			fprintf(stderr, "Invalid arguments\n");
			goto err;
	}

err:
	pthread_exit(NULL);
}


static void *do_j1939_test(void *args)
{
	int ret = 0;
	int i,j;
	int can_index;
	int can_amount;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_j1939_test on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X.%02X.%02X\n", 
			can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
	}

	// change mode to J1939 for all J1939 testing
	ret = AZ_VC_ModeActive(0x00, 0x02);
	if(ret != 0) {
		fprintf(stderr, "Calling port0 AZ_VC_ModeActive fail\n");
		goto err;
	}

	ret = AZ_VC_ModeActive(0x01, 0x02);
	if(ret != 0) {
		fprintf(stderr, "Calling port1 AZ_VC_ModeActive fail\n");
		goto err;
	}

	switch(j1939_test_arg) {
		case J1939_TEST_IDX_RW:
			{
				struct J1939_message_t msg;
				uint32_t read_size = 0;
				int port0_recv_count = 0,port1_recv_count = 0;

				j1939_send_msg.p = (j1939_send_msg.pgn >> 26 & 0x07);
				j1939_send_msg.sa = (j1939_send_msg.pgn & 0xFF);

				for(i=0; i<test_cnt; i++) {

					ret = AZ_VC_J1939_Write(j1939_send_msg);
					if(ret != 0) {
						fprintf(stderr, "Calling AZ_VC_J1939_Write fail\n");
						goto err;
					}				

					memset(&msg, 0x00, sizeof(msg));
					can_amount = AZ_VC_J1939_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						ret = AZ_VC_J1939_Read(&msg, &read_size);
						if(ret == 0)
						{
							
							fprintf(stdout, "Port0 receive count = %d\n", port0_recv_count +1);
							//fprintf(stdout, "J1939 read size = %d\n", read_size);
							fprintf(stdout, "\tid = 0x%08X\n", msg.id);
							fprintf(stdout, "\tp = 0x%02X\n", msg.p);
							fprintf(stdout, "\tdp = 0x%02X\n", msg.dp);
							fprintf(stdout, "\tpf = 0x%02X\n", msg.pf);
							fprintf(stdout, "\tps = 0x%02X\n", msg.ps);
							fprintf(stdout, "\tsa = 0x%02X\n", msg.sa);
							fprintf(stdout, "\tpgn= 0x%02X\n", msg.pgn);
							fprintf(stdout, "\tdlc = 0x%02X\n", msg.dlc);

							for(j=0; j<8; i++) {
								fprintf(stdout, "\tdata[%d] = 0x%02X\n", j, msg.data[j]);
							}
							port0_recv_count++;
						}
					}
					memset(&msg, 0x00, sizeof(msg));
					
					can_amount = AZ_VC_J1939_Read_CAN2_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{					
						ret = AZ_VC_J1939_Read_CAN2(&msg, &read_size);
						if(ret == 0)
						{
							fprintf(stdout, "Port1 Receive count = %d\n",port1_recv_count + 1);
							
							//fprintf(stdout, "J1939 read size = %d\n", read_size);
							fprintf(stdout, "\tid = 0x%08X\n", msg.id);
							fprintf(stdout, "\tp = 0x%02X\n", msg.p);
							fprintf(stdout, "\tdp = 0x%02X\n", msg.dp);
							fprintf(stdout, "\tpf = 0x%02X\n", msg.pf);
							fprintf(stdout, "\tps = 0x%02X\n", msg.ps);
							fprintf(stdout, "\tsa = 0x%02X\n", msg.sa);
							fprintf(stdout, "\tpgn= 0x%02X\n", msg.pgn);
							fprintf(stdout, "\tdlc = 0x%02X\n", msg.dlc);

							for(j=0; j<8; j++) {
								fprintf(stdout, "\tdata[%d] = 0x%02X\n", j, msg.data[j]);
							}
							port1_recv_count++;
						}
					}					
				}
					
			}
			break;

		case J1939_TEST_IDX_PORT_CONFIG:
			{
				ret = AZ_VC_J1939Port_Config(j1939_test_port);
				fprintf(stdout, "J1939 port config, ret = %d\n", ret);
			}
			break;

		case J1939_TEST_IDX_FILTER_CONFIG:
			{
				ret = AZ_VC_J1939Filter_Config(j1939_filter_config.type, j1939_filter_config.port, j1939_filter_config.pgn);
				fprintf(stdout, "J1939 filter config, ret = %d\n", ret);
			}
			break;

		case J1939_TEST_IDX_GET_FILTER_COUNTER:
			{
				uint16_t count = 0;

				AZ_VC_J1939_get_filter_count(j1939_test_port, &count);
				fprintf(stdout, "J1939 filter count = %d\n", count);
			}
			break;

		case J1939_TEST_IDX_FILTER_GET:
			{
				uint16_t count = 0;
				uint32_t *filter = NULL;

				AZ_VC_J1939_get_filter_count(j1939_test_port, &count);
				fprintf(stdout, "J1939 filter count = %d\n", count);

				if(count > 0) {
					int i;

					filter = malloc(sizeof(uint32_t) * count);
					memset(filter, 0x00, sizeof(uint32_t) * count);

					AZ_VC_J1939Filter_Get(j1939_test_port, filter);

					for(i=0; i<count; i++) {
						fprintf(stdout, "filter[%d] = 0x%08X\n", i, filter[i]);
					}

					free(filter);
				}
			}
			break;

		case J1939_TEST_IDX_READ:
			{
				struct J1939_message_t msg;
				struct J1939_message_t msg_can2;
				uint32_t read_size = 0;
				int port0_count = 0;
				int port1_count = 0;

				while(1)
				{
					if(stop_thread != 0)
						break;
					memset(&msg, 0x00, sizeof(msg));
					can_amount = AZ_VC_J1939_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						ret = AZ_VC_J1939_Read(&msg, &read_size);
						if(ret == 0)
						{
							fprintf(stdout, "Can port 0 receive= %d\n", port0_count + 1);					
							fprintf(stdout, "J1939 read size = %d\n", read_size);
							fprintf(stdout, "\tid = 0x%08X\n", msg.id);
							fprintf(stdout, "\tp = 0x%02X\n", msg.p);
							fprintf(stdout, "\tdp = 0x%02X\n", msg.dp);
							fprintf(stdout, "\tpf = 0x%02X\n", msg.pf);
							fprintf(stdout, "\tps = 0x%02X\n", msg.ps);
							fprintf(stdout, "\tsa = 0x%02X\n", msg.sa);
							fprintf(stdout, "\tpgn= 0x%02X\n", msg.pgn);
							fprintf(stdout, "\tdlc = 0x%02X\n", msg.dlc);

							for(i=0; i<8; i++) {
								fprintf(stdout, "\tdata[%d] = 0x%02X\n", i, msg.data[i]);
							}
							port0_count++;
						}
					}
					memset(&msg_can2, 0x00, sizeof(msg));
					
					can_amount = AZ_VC_J1939_Read_CAN2_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						ret = AZ_VC_J1939_Read_CAN2(&msg_can2, &read_size);
						if(ret == 0)
						{
							fprintf(stdout, "Can port 1 receive= %d\n", port1_count + 1);										

							
							fprintf(stdout, "J1939 read size = %d\n", read_size);
							fprintf(stdout, "\tid = 0x%08X\n", msg_can2.id);
							fprintf(stdout, "\tp = 0x%02X\n", msg_can2.p);
							fprintf(stdout, "\tdp = 0x%02X\n", msg_can2.dp);
							fprintf(stdout, "\tpf = 0x%02X\n", msg_can2.pf);
							fprintf(stdout, "\tps = 0x%02X\n", msg_can2.ps);
							fprintf(stdout, "\tsa = 0x%02X\n", msg_can2.sa);
							fprintf(stdout, "\tpgn= 0x%02X\n", msg_can2.pgn);
							fprintf(stdout, "\tdlc = 0x%02X\n", msg_can2.dlc);

							for(i=0; i<8; i++) {
								fprintf(stdout, "\tdata[%d] = 0x%02X\n", i, msg_can2.data[i]);
							}
												
							port1_count++;
						}
					}
				}
			}
			break;

		default:
			fprintf(stderr, "Invalid arguments\n");
			goto err;
	}

err:
	pthread_exit(NULL);
}

static void print_j1708_21_pkt(struct J1708_21_message_t *msg_21)
{
	fprintf(stdout, "\tpkt_no = 0x%02X,", msg_21->pkt_no);
	fprintf(stdout, "\tmid = 0x%02X,", msg_21->mid);
	fprintf(stdout, "\tp = 0x%02X,", msg_21->p);
	fprintf(stdout, "\tdlc = 0x%02X\n", msg_21->dlc);

	for(uint32_t i=0; i<sizeof(msg_21->data); i++) {
		if(i == 0)
		{
			fprintf(stdout, "\t");
		}
		if(i == sizeof(msg_21->data) - 1)
		{
			fprintf(stdout, " data[%d] = 0x%02X\n", i, msg_21->data[i]);
		}
		else
		{
			fprintf(stdout, " data[%d] = 0x%02X,", i, msg_21->data[i]);
		}
	}
}

static void *do_j1708_demo(void *args)
{
	int ret = 0;
	int j1708_index;
	int j1708_amount;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_demo_test on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	//get_cfv();

	switch(j1708_21_demo_arg) {
		case DEMO_TEST_IDX_J1708_READ_21:
			{
				struct J1708_message_t msg_27;
				struct J1708_21_message_t msg_21;
				struct J1708_21_message_t msg_21_1st;
				uint32_t read_size = 0;
				int read_21_packet_cnt = 0;

				while(stop_thread == 0) {
					{
						read_size = 0;
						memset(&msg_21, 0x00, sizeof(msg_21));
						
						j1708_amount = AZ_VC_J1708_21_Read_Amount();
						for(j1708_index = 0; j1708_index < j1708_amount; j1708_index++)
						{
							if(AZ_VC_J1708_21_Read(&msg_21, &read_size) == 0) {
								fprintf(stdout, "J1708 21 read size = %d, packet_cnt = %d\n", read_size, (read_21_packet_cnt + 1));
								print_j1708_21_pkt(&msg_21);

								if((msg_21.pkt_no == 0) && (msg_21.dlc > 13)) 
								{
									memcpy(&msg_21_1st, &msg_21, sizeof(msg_21_1st));
								}

								if((msg_21.pkt_no == 1) && (msg_21.dlc > 13)) {
									}
								read_21_packet_cnt++;
							}
						}
					}

					usleep(1);
				}
			}
			break;

		case DEMO_TEST_IDX_J1708_WRITE_21:
			{
				int cnt = 1;
				
				for(uint8_t i=0; i<test_cnt; i++) {
					if(stop_thread != 0)
						break;
					if(j1708_21_send_msg.dlc > 13)
					{
						memcpy(&j1708_21_send_msg.data[0], &j1708_send_allmsg.alldata[0], 13);
						
						j1708_21_send_msg.pkt_no = 0;
						AZ_VC_J1708_21_Write(j1708_21_send_msg);
						sleep(1);
						memcpy(&j1708_21_send_msg.data[0], &j1708_send_allmsg.alldata[13], 7);
						j1708_21_send_msg.pkt_no = 1;
						AZ_VC_J1708_21_Write(j1708_21_send_msg);
						
					}
					else
					{
						memcpy(&j1708_21_send_msg.data[0], &j1708_send_allmsg.alldata[0], 13);
						ret = AZ_VC_J1708_21_Write(j1708_21_send_msg);
					}
					if(ret != 0) {
						fprintf(stderr, "Calling AZ_VC_J708_21_Send fail\n");
						goto err;
					} else {
						fprintf(stdout, "cnt = %d\n",cnt);
						cnt++;
					}
				}
				stop_thread = 1;
			}
			break;

		case J1708_TEST_IDX_FILTER_CONFIG:			
			{				
				ret = AZ_VC_J1708_21byte_FilterConfig(j1708_filter_config.type, j1708_filter_config.mid);				
				fprintf(stdout, "J1708 filter config, ret = %d\n", ret);			
			}			
			break;		
			
		case J1708_TEST_IDX_GET_FILTER_COUNTER:			
			{				
				uint16_t count = 0;				
				AZ_VC_J1708_21byte_GetFilterCount(&count);				
				fprintf(stdout, "J1708 filter count = %d\n", count);			
			}			
			break;		
			
		case J1708_TEST_IDX_FILTER_GET:			
			{				
				uint16_t count = 0;				
				uint32_t *filter = NULL;				
				AZ_VC_J1708_21byte_GetFilterCount(&count);				
				fprintf(stdout, "J1708 filter count = %d\n", count);
				
				if(count > 0) 
				{					
					int i;					
					filter = malloc(sizeof(uint32_t) * count);					
					memset(filter, 0x00, sizeof(uint32_t) * count);					
					AZ_VC_J1708_21byte_FilterGet(filter);

					//fprintf(stdout, "filter = %d\n", filter);
					
					for(i=0; i<count; i++) {						
						fprintf(stdout, "filter[%d] = 0x%08X\n", i, filter[i]);					
						}					
					free(filter);				
				}			
			}			
			break;
			
		default:
			fprintf(stderr, "Invalid arguments\n");
			goto err;
	}

err:
	pthread_exit(NULL);
}

static void *do_ecompass_demo(void *args)
{
	int ret = 0;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_demo_test on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	struct ecompass_scale get_scale;

	switch(ecompass_demo_arg) {
		case DEMO_TEST_IDX_ecompass_set_accelerator:
			{
				ret = AZ_VC_set_accelerator(ecompass_acc_scale);
				if(ret != 0) {
					fprintf(stderr, "Calling AZ_VC_set_accelerator fail\n");
					goto err;
				} else {
					fprintf(stdout, "Calling AZ_VC_set_accelerator done\n");
				}
			}
			break;

		case DEMO_TEST_IDX_ecompass_get_accelerator:
			{
				ret = AZ_VC_get_accelerator(&get_scale);
				if(ret == -1) {
					fprintf(stderr, "Calling AZ_VC_get_accelerator fail\n");
					goto err;
				} else {
					fprintf(stdout, "Calling AZ_VC_get_accelerator success\n");
					fprintf(stdout, "scale value = ");
					switch(get_scale.scale)
					{
					case 0:
						fprintf(stdout, "2\n");
						break;
					case 1:
						fprintf(stdout, "4\n");
						break;
					case 2:
						fprintf(stdout, "8\n");
						break;
					case 3:
						fprintf(stdout, "16\n");
						break;
					}
				}
			}
			break;
		default:
			fprintf(stderr, "Invalid arguments\n");
			goto err;
	}

err:
	pthread_exit(NULL);
}

static void *do_algorithm_demo(void *args)
{
	int ret = 0;
	struct F_B_Algorithm_t f_b_algorithm;

	memset(&f_b_algorithm, 0x00, sizeof(f_b_algorithm));

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_demo_test on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	switch(algorithm_demo_arg) {
		case DEMO_TEST_IDX_ALGORITHM_CLEAR_G0Gf:
			{
				ret = AZ_VC_clear_sensor_algorithm();
				if(ret != 0) {
					fprintf(stderr, "Calling AZ_VC_clear_sensor_algorithm fail\n");
					goto err;
				} else {
					fprintf(stdout, "Calling AZ_VC_clear_sensor_algorithm done\n");
				}
			}
			break;
		case DEMO_TEST_IDX_ALGORITHM_SET_F_B_SWH:
			{
				ret = AZ_VC_Set_F_B_Algorithm(algorithm_enable);
				if(ret == -1) {
					fprintf(stderr, "Calling AZ_VC_Set_F_B_Algorithm fail\n");
					goto err;
				} else {
					fprintf(stdout, "Calling AZ_VC_Set_F_B_Algorithm success\n");
				}
			}
			break;
		case DEMO_TEST_IDX_ALGORITHM_GET_F_B_SWH:
			{
				if((ret = AZ_VC_Get_F_B_Algorithm(&f_b_algorithm)) != 0) {
					fprintf(stderr, "Calling AZ_VC_Get_F_B_Algorithm fail\n");
					ret = -1;
					goto err;
				}

				fprintf(stdout,"Calling AZ_VC_Get_F_B_Algorithm success\nStatus : ");
				switch(f_b_algorithm.status){
					case 0:
						fprintf(stdout,"Disable");
						break;
					case 1:
						fprintf(stdout,"Enable");
						break;
				}
				fprintf(stdout,".\n");
				fprintf(stdout,"Calibration : ");
				switch(f_b_algorithm.calibration){
					case 0:
						fprintf(stdout,"No Calibration");
						break;
					case 1:
						fprintf(stdout,"1 Calibration");
						break;
					case 2:
						fprintf(stdout,"2 Calibration");
						break;
					case 3:
						fprintf(stdout,"3 Calibration");
						break;
					case 4:
						fprintf(stdout,"4 Calibration");
						break;
					case 5:
						fprintf(stdout,"5 Calibration");
						break;
					default:
						fprintf(stdout,"NULL");
						break;
				}
				fprintf(stdout,".\n");
				stop_thread = 1;
			}
			break;

		default:
			fprintf(stderr, "Invalid arguments\n");
			goto err;
	}
	stop_thread = 1;

err:
	pthread_exit(NULL);
}

static void *do_set_store_config(void *argc)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_read_write_raw on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;
	
	ret = AZ_VC_Save_Config(store_config_mode);
	
	if(ret == 0)
	{
		fprintf(stdout, "Save config success.\n");
	}
	else
	{
		fprintf(stdout, "Save config fail.\n");
		goto err;
	}
	
	stop_thread = 1;
	
err:
	pthread_exit(NULL);
}

static void *do_set_request_time(void *argc)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_read_write_raw on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;
	
	ret = AZ_VC_Set_Request(set_request_time_port,set_request_time_time);
	
	if(ret == 0)
	{
		fprintf(stdout, "Set request success.\n");
	}
	else
	{
		fprintf(stdout, "Set request fail.\n");
		goto err;
	}
	
	stop_thread = 1;
	
err:
	pthread_exit(NULL);
}

static void *do_read_sku(void *argc)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_read_write_raw on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;
	
	ret = AZ_VC_read_sku(&module_sku);
	
	if(ret == 0)
	{

		fprintf(stdout, "read sku success.\n");
		fprintf(stdout, "sku = %02X%02X%02X%02X%02X%02X%02X%02X\n", 
			module_sku.sku1,module_sku.sku2,module_sku.sku3,module_sku.sku4,module_sku.sku5,module_sku.sku6,module_sku.sku7,module_sku.sku8);
	}
	else
	{
		fprintf(stdout, "Set request fail.\n");
		goto err;
	}
	
	stop_thread = 1;
	
err:
	pthread_exit(NULL);
}

static int do_functions(void)
{
	int ret = 0;
	pthread_t deinit_wait_thread;
	pthread_t do_get_can_fw_ver_thread;
	pthread_t do_can_config_thread;
	pthread_t do_mode_active_thread;
	pthread_t do_filter_config_thread;
	pthread_t do_filter_config_raw_thread;
	pthread_t do_get_filter_config_thread;
	pthread_t do_rawcan_demo_thread;
	pthread_t do_launch_receiver_thread;
	pthread_t do_j1939_test_thread;
	pthread_t do_reset_module_thread;
	pthread_t do_get_uart_canbaudrate_thread;
	pthread_t do_set_uart_baudrate_thread;
	pthread_t do_set_store_config_thread;
	pthread_t do_set_request_time_thread;
	pthread_t do_j1708_demo_thread;
	pthread_t do_ecompass_demo_thread;
	pthread_t do_algorithm_demo_thread;
	pthread_t do_read_sku_thread;

	fprintf(stdout, "version %d.%d.%d\n",
		FARO_CAN_SDK_DEMO_VERSION_MAJOR,
		FARO_CAN_SDK_DEMO_VERSION_MINOR,
		FARO_CAN_SDK_DEMO_VERSION_PATCH);

	if(scan_port) {
		if((ret = do_scan_port()) != 0) {
			fprintf(stderr, "do_scan_port fail\n");
		}

		goto err;
	}

	// spawn deinit_wait_thread if necessary
	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "spawning deinit_wait thread...\n");

	if((ret = pthread_create(&deinit_wait_thread, NULL, deinit_wait, NULL)) != 0) {
		fprintf(stderr, "create deinit_wait thread fail\n");
			goto err;
	}

	if(get_can_fw_ver) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_get_can_fw_ver thread...\n");

		if((ret = pthread_create(&do_get_can_fw_ver_thread, NULL, do_get_can_fw_ver, NULL)) != 0) {
			fprintf(stderr, "create do_get_can_fw_ver thread fail\n");
			goto err;
		}

		ret = pthread_join(do_get_can_fw_ver_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_get_can_fw_ver_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}
	
	if(get_uart_canbaudrate) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_get_uart_canbaudrate thread...\n");

		if((ret = pthread_create(&do_get_uart_canbaudrate_thread, NULL, do_get_uart_canbaudrate, NULL)) != 0) {
			fprintf(stderr, "create do_get_uart_canbaudrate thread fail\n");
			goto err;
		}

		ret = pthread_join(do_get_uart_canbaudrate_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_get_uart_canbaudrate_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}
	
	if(set_uart_baudrate) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_set_uart thread...\n");

		if((ret = pthread_create(&do_set_uart_baudrate_thread, NULL, do_set_uart_baudrate, NULL)) != 0) {
			fprintf(stderr, "create do_set_uart_baudrate thread fail\n");
			goto err;
		}

		ret = pthread_join(do_set_uart_baudrate_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_set_uart_canbaudrate_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}
	
	if(reset_module) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_reset_module thread...\n");

		if((ret = pthread_create(&do_reset_module_thread, NULL, do_reset_module, NULL)) != 0) {
			fprintf(stderr, "create do_reset_module thread fail\n");
			goto err;
		}

		ret = pthread_join(do_reset_module_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_reset_module_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(can_config) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_can_config thread...\n");
		if((ret = pthread_create(&do_can_config_thread, NULL, do_can_config, NULL)) != 0) {
			fprintf(stderr, "create do_can_config thread fail\n");
			goto err;
		}

		ret = pthread_join(do_can_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_can_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(mode_active) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_mode_active thread...\n");
		if((ret = pthread_create(&do_mode_active_thread, NULL, do_mode_active, NULL)) != 0) {
			fprintf(stderr, "create do_mode_active thread fail\n");
			goto err;
		}

		ret = pthread_join(do_mode_active_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_mode_active_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(filter_config) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_filter_config thread...\n");
		if((ret = pthread_create(&do_filter_config_thread, NULL, do_filter_config, NULL)) != 0) {
			fprintf(stderr, "create do_filter_config thread fail\n");
			goto err;
		}

		ret = pthread_join(do_filter_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_filter_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	} 

	if(filter_config_raw) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_filter_config_raw thread...\n");
		if((ret = pthread_create(&do_filter_config_raw_thread, NULL, do_filter_config_raw, &can_filter_config)) != 0) {
			fprintf(stderr, "create do_filter_config_raw thread fail\n");
			goto err;
		}

		ret = pthread_join(do_filter_config_raw_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_filter_config_raw_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(get_filter_config) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_get_filter_config thread...\n");
		if((ret = pthread_create(&do_get_filter_config_thread, NULL, do_get_filter_config, NULL)) != 0) {
			fprintf(stderr, "create do_get_filter_config thread fail\n");
			goto err;
		}

		ret = pthread_join(do_get_filter_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_get_filter_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(rawcan_demo){
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_rawcan_demo thread...\n");
		if((ret = pthread_create(&do_rawcan_demo_thread, NULL, do_rawcan_demo, NULL)) != 0) {
			fprintf(stderr, "create do_rawcan_demo thread fail\n");
			goto err;
		}

		ret = pthread_join(do_rawcan_demo_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_rawcan_demo_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(j1939_test) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn do_j1939_test thread...\n");
		if((ret = pthread_create(&do_j1939_test_thread, NULL, do_j1939_test, NULL)) != 0) {
			fprintf(stderr, "create do_j1939_test thread fail\n");
			goto err;
		}

		ret = pthread_join(do_j1939_test_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_j1939_test error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(j1708_21_demo)
	{
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn do_j1708_demo thread...\n");
		if((ret = pthread_create(&do_j1708_demo_thread, NULL, do_j1708_demo, NULL)) != 0) {
			fprintf(stderr, "create do_j1708_demo thread fail\n");
			goto err;
		}

		ret = pthread_join(do_j1708_demo_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_j1708_demo error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(ecompass_demo)
	{
		do_ecompass_demo_thread;
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn do_ecompass_demo thread...\n");
		if((ret = pthread_create(&do_ecompass_demo_thread, NULL, do_ecompass_demo, NULL)) != 0) {
			fprintf(stderr, "create do_ecompass_demo thread fail\n");
			goto err;
		}

		ret = pthread_join(do_ecompass_demo_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_ecompass_demo error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}
	
	if(store_config)
	{
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn do_set_store_config_thread...\n");
		if((ret = pthread_create(&do_set_store_config_thread, NULL, do_set_store_config, NULL)) != 0) {
			fprintf(stderr, "create do_set_store_config_thread fail\n");
			goto err;
		}

		ret = pthread_join(do_set_store_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_set_store_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}
	
	if(set_request)
	{
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn set_request_time_thread...\n");
		if((ret = pthread_create(&do_set_request_time_thread, NULL, do_set_request_time, NULL)) != 0) {
			fprintf(stderr, "create do_set_request_time_thread fail\n");
			goto err;
		}

		ret = pthread_join(do_set_request_time_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_set_request_time error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(algorithm_demo)
	{
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn set_algorithm_demo_thread...\n");
		if((ret = pthread_create(&do_algorithm_demo_thread, NULL, do_algorithm_demo, NULL)) != 0) {
			fprintf(stderr, "create do_algorithm_demo_thread fail\n");
			goto err;
		}

		ret = pthread_join(do_algorithm_demo_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_algorithm_demo error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(read_sku)
	{
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn read_sku_thread...\n");
		if((ret = pthread_create(&do_read_sku_thread, NULL, do_read_sku, NULL)) != 0) {
			fprintf(stderr, "create read_sku_thread fail\n");
			goto err;
		}

		ret = pthread_join(do_read_sku_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_read_sku error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

normal_exit:
	ret = pthread_join(deinit_wait_thread, NULL);
	if(ret) {
		fprintf(stderr, "deinit_wait thread error: %s\n", strerror(ret));
		goto err;
	}


	fprintf(stderr, "threads are stopped\n");
	fflush(stderr);

err:
	scan_port = 0;
	get_can_fw_ver = 0;
	can_config = 0;
	mode_active = 0;
	filter_config = 0;
	filter_config_raw = 0;
	get_filter_config = 0;
	read_write_raw = 0;
	rawcan_demo = 0;
	j1939_test = 0;
	set_uart_baudrate = 0;
	get_uart_canbaudrate = 0;
	store_config = 0;
	set_request = 0;
	j1708_21_demo = 0;
	ecompass_demo = 0;
	algorithm_demo = 0;
	read_sku = 0;
	return ret;
}

int main(int argc, char *argv[])
{
	int ret;

	if(signal(SIGINT, sig_handler) == SIG_ERR)
		fprintf(stderr, "fail to catch SIGINT\n");

	if(process_options(argc, argv) == 0)
		do_functions();

    if(port_name != NULL)
    	free(port_name);

	return 0;
}
