/***********************************************************************
 *
 *  Copyright (c) 2019 - 2020 ANTZER TECH CO., LTD.
 *  All Rights Reserved
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <linux/limits.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "faro_can_sdk_demo_config.h"
#include "faro_can_sdk.h"

#define fprintf(fmt,...) {if (!g_quiet_print_flag) printf(__VA_ARGS__);}

#define FARO_CAN_SDK_OBD2_WRITE								8
#define FARO_CAN_SDK_OBD2_READ								1
#define FARO_CAN_SDK_OBD2_READ_Sample						3
#define FARO_CAN_SDK_OBD2_READ_Sample_DECODE				3

#define FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_32_BITS		6
#define FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_16_BITS		8
#define FARO_CAN_SDK_FILTER_CONFIG_RESET_REMOVE			4
#define FARO_CAN_SDK_CAN_CONFIG_ARGS					2
#define FARO_CAN_SDK_GET_FILTER_CONFIG					2

enum OBD2_TEST_IDX
{
	OBD2_TEST_IDX_READ = 0,
	OBD2_TEST_IDX_WRITE,
	OBD2_TEST_IDX_WRITE_sample_PID,
	OBD2_TEST_IDX_WRITE_sample_PID_DECODE,
	OBD2_TEST_IDX_WRITE_sample_PID_EXT,
	OBD2_TEST_IDX_WRITE_sample_PID_DECODE_EXT,
	NUM_OF_OBD_CNT_PKT,
};

static struct OBD2_send_msg_t obd2_send_msg;
static struct can_config_t can_config_data;
static struct can_filter_config_t can_filter_config;

static int deinit_in_seconds = (10 * 3);	//set deinit time
static int stop_thread = 0;
static char *port_name = NULL;
static uint8_t get_can_fw_ver = 0;
static uint8_t scan_port = 0;
static uint8_t can_config = 0;
static int test_cnt = 1;
static uint8_t mode_active = 0;
static int mode_active_port = 0;
static int mode_active_arg = 1;
static uint8_t filter_config = 0;
static uint8_t filter_config_raw = 0;
static uint8_t get_filter_config = 0;

static uint8_t obd2_test = 0;
static int obd2_test_arg = 0;
static int obd2_test_sample_cnt = 0;

static void display_help(int argc, char *argv[])
{
	fprintf(stdout, "Usage: %s [OPTION]\n"
			"\n"
			"====  Using OBD2 function, need to do ModeActive to OBD2 mode first  ====\n"
			"  -h, --help\n"
			"  -s, --scan, scan available serial port\n"
			"  -p, --port\n"
			"		#1-port, ex \"/dev/ttyUSB0\"\n"
			"  -d, --deinit\n"
			"		#1-deinit after x seconds, -1 endless until ctrl-c\n"
			"  -g, --get_fw_ver, get CAN firmware version\n"
			"  -c, --can_config, can config\n"
			"		#1-port: 1 byte\n"
			"		#2-speed: 1 byte, 0: 1M, 1: 800K, 2: 500K, 3: 250K, 4: 200K, 5: 125K\n"
			"  -m, --mode_active\n"
			"		#1-port: 0 or 1\n"
			"		#2-mode active: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR_OBD2, 5:CAN2ADR_J1939, 8:J1708_21, 9: J1708_21_OFF\n"
			"  -w, --filter_config_raw\n"
			"		#1-type: 0: ID+Mask, 1: ID List, 2: Remove, 3: Reset All\n"
			"		#2-port: 0, 1\n"
			"		#3-bank: 0 - 13\n"
			"		#4-mode: 0: 16 bits, 1: 32 bits\n"
			"		#5-filterId: 16/32 bits value in hex, ex: 0x0011/0x00112233\n"
			"		#6-filterId/filterMask: 16bits/32 bits in hex, ex: 0x1234/0x00112233\n"
			"		#7-filterMask: 16 bits in hex, ex: 0x0011\n"
			"		#8-filterMask: 16 bits in hex, ex: 0x0011\n"
			"  -x, --get_filter_config\n"
			"       #1 port: 0, 1\n"
			"       #2 bank: 0 - 13\n"
			"  -o, --obd2_test\n"
			"		0 - obd2 read\n"
			"		1 - obd2 write and receive data\n"
			"			#1 -port: 0 or 1\n"
			"			#2 -Can ID: 0x7DF ~ 0x7E7\n"
			"			#3 -Can Mode : 0 - CAN 2.0A, 1 - CAN 2.0B\n"
			"			#4 -DLC: 0 ~ 8\n"
			"			#5 -PID: (1byte)\n"
			"			#6 -Data(4byte) , ex: 0x01020304\n"
			"			#7 -Data(4byte) , ex: 0x05060708\n"
			"		2 - send obd2 PID sample and receive data\n"
			"			#1 -port: 0 or 1\n"			
			"			#2 -PID select:\n"
			"				\t1.  0x0D - (VSS)Vehicle Speed\n"
			"				\t2.  0x0C - (RPM)Vehicle RPM\n"
			"				\t3.  0x2F - (FLI)Fuel Tank Level Input\n"
			"				\t4.  0x46 - (AAT)Ambient air temperature\n"
			"				\t5.  0x5E - (EFR)Engine fuel rate\n"
			"				\t6.  0x42 - (VEP)Control module voltage\n"
			"				\t7.  0x04 - (PCT)Calculated engine load\n"
			"				\t8.  0x05 - (ECT)Engine coolant temperature\n"
			"				\t9.  0x10 - (MAF)Mass Air Flow\n"
			"				\t10. 0x45 - (ETP)Throttle Position\n"
			"				\t11. 0x33 - (ABP)Absolute Barometric Pressure\n"
			"				\t12. 0x0F - (IAT)Intake air temperature\n"
			"				\t13. 0x0A - (ERP)Fuel pressure\n"
			"				\t14. 0x5C - (EOT1)Engine oil temperature\n"
			"				\t15. 0x5A - (APP)Relative Accelerator Pedal Position\n"
			"				\t16. 0x02 - (VIN)Vehicle identfiflcation Number\n"
			"		3 - send obd2 PID sample and receive decode data\n"
			"			#1 -port: 0 or 1\n"			
			"			#2 -PID select:\n"
			"				\t1.  0x0D - (VSS)Vehicle Speed\n"
			"				\t2.  0x0C - (RPM)Vehicle RPM\n"
			"				\t3.  0x2F - (FLI)Fuel Tank Level Input\n"
			"				\t4.  0x46 - (AAT)Ambient air temperature\n"
			"				\t5.  0x5E - (EFR)Engine fuel rate\n"
			"				\t6.  0x42 - (VEP)Control module voltage\n"
			"				\t7.  0x04 - (PCT)Calculated engine load\n"
			"				\t8.  0x05 - (ECT)Engine coolant temperature\n"
			"				\t9.  0x10 - (MAF)Mass Air Flow\n"
			"				\t10. 0x45 - (ETP)Throttle Position\n"
			"				\t11. 0x33 - (ABP)Absolute Barometric Pressure\n"
			"				\t12. 0x0F - (IAT)Intake air temperature\n"
			"				\t13. 0x0A - (ERP)Fuel pressure\n"
			"				\t14. 0x5C - (EOT1)Engine oil temperature\n"
			"				\t15. 0x5A - (APP)Relative Accelerator Pedal Position\n"
			"				\t16. 0x02 - (VIN)Vehicle identfiflcation Number\n"
			"  -t, --test_cnt, set counter for testing loop\n"
			"  -v, --version\n"
			"  -q, --quiet_mode, set 1 to quiet mode\n"
			"\n",
			argv[0]);
	exit(0);
}

static int process_options(int argc, char *argv[])
{
	unsigned int tmp = 0;

	if(argc < 2) {
		display_help(argc, argv);
		return -1;
	}

	for (;;) {
		int option_index = 0;
		static const char *short_options = "hp:sd:gc:bm:fw:x:rj:lt:vuU:S:a:7:e:A:k:o:q:";
		static const struct option long_options[] = {
			{"help", no_argument, 0, 'h'},
			{"port", required_argument, 0, 'p'},
			{"scan", no_argument, 0, 's'},
			{"deinit", required_argument, 0, 'd'},
			{"get_fw_ver", no_argument, 0, 'g'},
			{"can_config", required_argument, 0, 'c'},
			{"reset_module", no_argument, 0, 'b'},
			{"mode_active", required_argument, 0, 'm'},
			{"filter_config", no_argument, 0, 'f'},
			{"filter_config_raw", required_argument, 0, 'w'},
			{"get_filter_config", required_argument, 0, 'x'},
			{"read_write_raw", no_argument, 0, 'r'},
			{"j1939_test", required_argument, 0,'j'},
			{"obd2_test", required_argument, 0,'o'},
			{"test_cnt", required_argument, 0, 't'},
			{"version", no_argument, 0, 'v'},
			{"quiet_mode", required_argument, 0, 'q'},
			{"set_uart", required_argument, 0, 'U'},
			{"get_uart_canspeed", no_argument, 0, 'u'},
			{"store_config",required_argument,0,'S'},
			{"set_request_time",required_argument,0,'a'},
			{"demo_j1708",required_argument,0,'7'},
			{"ecompass",required_argument,0,'e'},
			{"Algorithm",required_argument,0,'A'},
			{"read_sku",no_argument,0,'k'},
			{0,0,0,0},
		};

		int c = getopt_long(argc, argv, short_options, long_options, &option_index);
		if (c == EOF) {
			break;
		}

		switch (c) {
		case 'h':
			display_help(argc, argv);
			break;

		case 'p':
			port_name = strdup(optarg);
			break;

		case 's':
			scan_port = 1;
			break;

		case 'd':
			deinit_in_seconds = atoi(optarg);
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "deinit after %d seconds\n", deinit_in_seconds);
			break;

		case 'g':
			get_can_fw_ver = 1;
			break;

		case 'c':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				can_config = 1;
				memset(&can_config_data, 0x00, sizeof(can_config_data));

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0:
							can_config_data.port = atoi(argv[index]);
							if(can_config_data.port < 0 || can_config_data.port > 1)
								goto c_err;

							break;

						case 1:
							can_config_data.speed = atoi(argv[index]);
							if(can_config_data.speed < 0 || can_config_data.speed > 5)
								goto c_err;

							break;

						default:
							goto c_err;
					}

					index++;
					cnt++;
				};

				arg_num = FARO_CAN_SDK_CAN_CONFIG_ARGS;

				if(cnt != arg_num)
					goto c_err;

				optind = index - 1;
				goto c_ok;

c_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

c_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -c arguments done\n");
			}
			break;

		case 'm':
		{
			int cnt = 0;
			int index = optind - 1;
			int arg_num = 0;
			while(index < argc) {
				if(argv[index][0] == '-') {
					break;
				}

				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);
				
				switch(cnt) {
					case 0:
						if(atoi(argv[index]) < 0 || atoi(argv[index]) > 1)
						{
							fprintf(stdout,
							"invalid argument number.\nPort: 0 or 1\nMode: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR OBD2, 5:CAN2ADR J1939, 8:J1708_21 9:J1708_21_OFF\n");
							return -1;
						}
						else
						{
							mode_active_port = atoi(argv[index]);
						}
						break;

					case 1:
						//if(atoi(argv[index]) < 0 || atoi(argv[index]) > 10)
						if(atoi(argv[index]) != 0 && atoi(argv[index]) != 1 && atoi(argv[index]) != 2 && atoi(argv[index]) != 3 && atoi(argv[index]) != 4 && atoi(argv[index]) != 5 && atoi(argv[index]) != 8 && atoi(argv[index]) != 9)
						{
							fprintf(stdout,
							"invalid argument number.\nPort: 0 or 1\nMode: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR OBD2, 5:CAN2ADR J1939, 8:J1708_21 9:J1708_21_OFF\n");
							return -1;
						}
						else
						{
							mode_active_arg = atoi(argv[index]);
							mode_active = 1;
						}
						break;

					//default:
						
				}
				
				index++;
				cnt++;
				
				arg_num = FARO_CAN_SDK_CAN_CONFIG_ARGS;
				
			}
			if(cnt != arg_num)
				goto m_err;

			optind = index - 1;
			goto m_ok;

m_err:
			fprintf(stderr, "invalid argument number, should be %d\n", arg_num);

m_ok:
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "checking -m arguments done\n");
			break;
		}
		
		case 'o':
		{
			int cnt = 0;
			int index = optind - 1;
			int arg_num = 0;

			obd2_test = 1;

			while(index < argc) {
				if(argv[index][0] == '-') {
					break;
				}

				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

				tmp = 0;

				switch(cnt) {
					case 0:
						obd2_test_arg = atoi(argv[index]);
						if(obd2_test_arg < 0 || obd2_test_arg > NUM_OF_OBD_CNT_PKT)
							goto o_err;

						if(obd2_test_arg != OBD2_TEST_IDX_READ) {
							memset(&obd2_send_msg, 0x00, sizeof(obd2_send_msg));
						} 
						break;

					case 1:
						if(obd2_test_arg != OBD2_TEST_IDX_READ) {
							//sscanf(argv[index], "0x%02X", &tmp);
							sscanf(argv[index], "%d", &tmp);
							obd2_send_msg.port = tmp;
						} 
						break;

					case 2:
						if(obd2_test_arg == OBD2_TEST_IDX_WRITE) {
							sscanf(argv[index], "0x%08X", &tmp);
							tmp = ntohl(tmp);
							obd2_send_msg.id= tmp;
						} else if (obd2_test_arg == OBD2_TEST_IDX_WRITE_sample_PID
						|| obd2_test_arg == OBD2_TEST_IDX_WRITE_sample_PID_DECODE) {
							//sscanf(argv[index], "%02X", &tmp);
							sscanf(argv[index], "%d", &tmp);
							//tmp = ntohl(tmp);
							obd2_test_sample_cnt = tmp;
							fprintf(stdout, "tmp = %d\n",obd2_test_sample_cnt);
						} else if (obd2_test_arg == OBD2_TEST_IDX_WRITE_sample_PID_EXT
						|| obd2_test_arg == OBD2_TEST_IDX_WRITE_sample_PID_DECODE_EXT) {
							//sscanf(argv[index], "%02X", &tmp);
							sscanf(argv[index], "%d", &tmp);
							//tmp = ntohl(tmp);
							obd2_test_sample_cnt = tmp;
							fprintf(stdout, "tmp = %d\n",obd2_test_sample_cnt);
						}
						break;

					case 3:
						if(obd2_test_arg == OBD2_TEST_IDX_WRITE) {
							sscanf(argv[index], "0x%02X", &tmp);
							obd2_send_msg.can_mode = tmp;
						}
						break;

					case 4:
						if(obd2_test_arg == OBD2_TEST_IDX_WRITE) {
							//sscanf(argv[index], "0x%02X", &tmp);
							sscanf(argv[index], "%d", &tmp);
							obd2_send_msg.dlc = tmp;
						}
						break;

					case 5:
						if(obd2_test_arg == OBD2_TEST_IDX_WRITE) {
							sscanf(argv[index], "0x%02X", &tmp);
							obd2_send_msg.PID = tmp;
						}
						break;

					case 6:
						if(obd2_test_arg == OBD2_TEST_IDX_WRITE) {
							sscanf(argv[index], "0x%08X", &tmp);
							tmp = ntohl(tmp);
							memcpy(&obd2_send_msg.data[0], &tmp, sizeof(uint32_t));
						}
						break;

					case 7:
						if(obd2_test_arg == OBD2_TEST_IDX_WRITE) {
							sscanf(argv[index], "0x%08X", &tmp);
							tmp = ntohl(tmp);
							memcpy(&obd2_send_msg.data[4], &tmp, sizeof(uint32_t));
						}
						break;

					default:
						goto o_err;
				}

				index++;
				cnt++;
			};
			switch(obd2_test_arg) {
				case OBD2_TEST_IDX_READ:		arg_num = FARO_CAN_SDK_OBD2_READ;				break;
				case OBD2_TEST_IDX_WRITE:		arg_num = FARO_CAN_SDK_OBD2_WRITE;			break;
				case OBD2_TEST_IDX_WRITE_sample_PID:	arg_num = FARO_CAN_SDK_OBD2_READ_Sample;		break;
				case OBD2_TEST_IDX_WRITE_sample_PID_DECODE: arg_num = FARO_CAN_SDK_OBD2_READ_Sample_DECODE; break;
				case OBD2_TEST_IDX_WRITE_sample_PID_EXT:	arg_num = FARO_CAN_SDK_OBD2_READ_Sample;		break;
				case OBD2_TEST_IDX_WRITE_sample_PID_DECODE_EXT: arg_num = FARO_CAN_SDK_OBD2_READ_Sample_DECODE; break;
			}

			if(cnt != arg_num)
				goto o_err;

			optind = index - 1;
			goto o_ok;

o_err:
			fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
			//display_help(argc, argv);

o_ok:
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "checking -j arguments done\n");
		}
		break;

		case 'w':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				filter_config_raw = 1;
				memset(&can_filter_config, 0x00, sizeof(can_filter_config));

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0: // 0 - ID+Mask, 1 - ID List, 2 - Remove, 3 - Reset All
							can_filter_config.type = atoi(argv[index]);
							if(can_filter_config.type < 0 || can_filter_config.type > 3)
								goto w_err;

							break;

						case 1:
							can_filter_config.port = atoi(argv[index]);
							if(can_filter_config.port < 0 && can_filter_config.port > 1)
								goto w_err;

							break;

						case 2:
							can_filter_config.bank = atoi(argv[index]);
							if(can_filter_config.bank < 0 || can_filter_config.bank > 13)
								goto w_err;

							break;

						case 3:
							can_filter_config.mode = atoi(argv[index]);
							if(can_filter_config.mode < 0 || can_filter_config.mode > 1)
								goto w_err;

							break;

						case 4:
							if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
								sscanf(argv[index], "0x%08X", &can_filter_config.filterId);
							} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterId |= (tmp << 16);
							}
							break;

						case 5:
							if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
								sscanf(argv[index], "0x%08X", &can_filter_config.filterMask);
							} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterId |= tmp;
							}
							break;

						case 6:
							if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterMask |= (tmp << 16);
							}
							break;

						case 7:
							if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterMask |= tmp;
							}
							break;

						default:
							goto w_err;
					}

					index++;
					cnt++;
				};

				if((can_filter_config.type == 2) || (can_filter_config.type == 3)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RESET_REMOVE;
				} else if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
					(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_32_BITS;
				} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
					(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_16_BITS;
				}

				if(cnt != arg_num)
					goto w_err;

				optind = index - 1;
				goto w_ok;

w_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

w_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -w arguments done\n");
			}
			break;

		case 't':
			test_cnt = atoi(optarg);
			if(test_cnt <= 0)
				test_cnt = 1;
			break;

		case 'v':
			fprintf(stdout, "%s Version %d.%d.%d\n",
				argv[0],
				FARO_CAN_SDK_DEMO_VERSION_MAJOR,
				FARO_CAN_SDK_DEMO_VERSION_MINOR,
				FARO_CAN_SDK_DEMO_VERSION_PATCH);
			return -1;

		case 'q':
			g_quiet_print_flag = atoi(optarg);
			sleep(1);
			if(g_quiet_print_flag > 1 || g_quiet_print_flag < 0)
			{
				fprintf(stdout, "invalid argument number.\n0:Normal mode, 1:Quiet mode\n");
				return -1;
			}
			break;
		}
	}

	return 0;
}

static void *deinit_wait(void *args)
{
	int i;

	if(deinit_in_seconds == -1) {
		while(stop_thread == 0)
			sleep(1);
	} else {
		for(i=0; i<deinit_in_seconds; i++) {
			if(stop_thread != 0)
				break;

			sleep(1);
		}
	}

	if(stop_thread == 0) {
		stop_thread = 1;
		sleep(1);
	}

	AZ_VC_DeInit();

	pthread_exit(NULL);
}

void sig_handler(int signo)
{
	if(signo == SIGINT) {
		fprintf(stderr, "catch SIGINT\n");
		stop_thread = 1;
	}
}

static int do_scan_port(void)
{
	int ret = 0;
	char buf[PATH_MAX];
	int serial_port_num = 0;

	memset(buf, sizeof(buf), 0x00);
	AZ_VC_GetCommPort(buf, &serial_port_num);
	fprintf(stdout, "Found %d serial ports\n", serial_port_num);
	if(serial_port_num > 0)
		fprintf(stdout, "buf -> \n%s", buf);

	return ret;
}

static void *do_get_can_fw_ver(void *args)
{
	int ret = 0;
	int i;
	struct can_fw_ver_t can_fw_ver;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting get firmware version on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail, test_cnt = %d\n", i);
			continue;
		}

		fprintf(stdout, "get FW count = %d, can_fw_ver: %02X.%02X.%02X.%02X.%02X\n",
				i+1, can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
		
		/*
		fprintf(stdout, "get FW count = %d, can_fw_ver: %02X.%02X.%02X\n",
				i+1, can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
		*/
	}

err:
	pthread_exit(NULL);
}

static void *do_can_config(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting can_config on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X.%02X.%02X\n", 
			can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		if((ret = AZ_VC_CAN_Config(can_config_data)) != 0)
			goto err;

		fprintf(stdout, "test_cnt = %d, port = %d, speed = %d\n",
			i, can_config_data.port, can_config_data.speed);
	}

err:
	stop_thread = 1;
	pthread_exit(NULL);
}

static void *do_filter_config_raw(void *args)
{
	int ret = 0;
	struct can_filter_config_t *filter = (struct can_filter_config_t *)args;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "type = %d, port = %d, bank = %d, mode = %d, filterId = 0x%08X, filerMask = 0x%08X\n",
			filter->type, filter->port, filter->bank, filter->mode, filter->filterId, filter->filterMask);

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_filter_config_raw on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

/*
	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}
*/

	if((ret = AZ_VC_CAN_Filter_Config(*filter)) != 0)
		goto err;
	else
	{
		
		fprintf(stdout, "Set filter config success.\n");
		stop_thread = 1;
	}

err:
	pthread_exit(NULL);
}

static void *do_get_filter_config(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_get_filter_config on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X.%02X.%02X\n", 
			can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type, can_fw_ver.year, can_fw_ver.week);
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		fprintf(stdout, "set port = %d, bank = %d\n",
			can_filter_config.port, can_filter_config.bank);

		if((ret = AZ_VC_CAN_Filter_Get(&can_filter_config)) != 0)
			goto err;

		fprintf(stdout, "test_cnt = %d:\n", i);
		fprintf(stdout, "\ttype = %d, port = %d, bank = %d, mode = %d\n",
			can_filter_config.type, can_filter_config.port, can_filter_config.bank, can_filter_config.mode);
		fprintf(stdout, "\tfilterId = 0x%08X, filterMask = 0x%08X\n",
			can_filter_config.filterId, can_filter_config.filterMask);
	}

err:
	pthread_exit(NULL);
}


static void *do_mode_active(void *args)
{
	int ret = 0;
	int i;
	static int current_active_mode = 0;

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		if(port_name == NULL) {
			fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
			goto err;
		}
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;
		
		if(mode_active_arg == 4 || mode_active_arg == 5)
		{
			fprintf(stdout, "Active ADR and checking...\n");
		}
		usleep(1);
		if((ret = AZ_VC_ModeActive(mode_active_port, mode_active_arg)) != 0) {
			fprintf(stderr, "calling AZ_VC_ModeActive fail\n");
			goto err;
		}

		fprintf(stdout, "test_cnt = %d, mode_active argument : %d success.\n", i, mode_active_arg);
	}
	stop_thread = 1;

err:
	pthread_exit(NULL);
}


static void *do_obd2_demo(void *args)
{
	uint8_t ret = 0;
	int can_index;
	int can_amount;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_demo_test on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	// change mode to J1939 for all J1939 testing
	ret = AZ_VC_ModeActive(0x00, 0x01);
	if(ret != 0) {
		fprintf(stderr, "Calling AZ_VC_ModeActive fail\n");
		goto err;
	}

	struct OBD2_message_t msg;

	switch(obd2_test_arg) {
		case OBD2_TEST_IDX_READ:
			{				
				uint32_t read_size = 0;
				int i, rec_cnt;
				while(1)
				{
					if(stop_thread !=0)
					{
						break;
					}
					ret = 0;
					can_amount = AZ_VC_OBD2_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						if(AZ_VC_OBD2_Read(&msg, &read_size) == 0)
						{
							fprintf(stdout, "OBD2 read size = %d", read_size);
							fprintf(stdout, "\treceive cnt = %d\n",rec_cnt + 1);
							fprintf(stdout, "port = %d", msg.port);
							if(msg.can_mode == 0)
							{
								fprintf(stdout, "\tID = 0x%03X", msg.id);
							}
							else
							{
								fprintf(stdout, "\tID = 0x%08X", msg.id);
							}
							fprintf(stdout, "\tservice = 0x%02X", msg.service);
							fprintf(stdout, "\tpid = 0x%02X", msg.pid);
							fprintf(stdout, "\tcan_mode = 0x%02X", msg.can_mode);
							fprintf(stdout, "\tdlc = 0x%02X\n", msg.dlc);
							for(i=0; i<8; i++) {
								fprintf(stdout, "data[%d] = 0x%02X\t", i, msg.data[i]);
							}
							fprintf(stdout, "\n");
							rec_cnt ++;
						}
					}
					//fprintf(stdout,"ret = %d\n",ret);
					usleep(1000);
				}
			}
			break;

		case OBD2_TEST_IDX_WRITE:
			{
				int i,j, k, rec_cnt;
				uint32_t read_size = 0;
				for(i = 0; i < test_cnt; i++)
				{
					if(stop_thread !=0)
					{
						break;
					}
					ret = AZ_VC_OBD2_Write(obd2_send_msg);
					if(ret == -1) {
						fprintf(stderr, "Calling AZ_VC_OBD2_Write fail\n");
						goto err;
					} else {
						fprintf(stdout, "Calling AZ_VC_OBD2_Write done\n");
					}
					usleep(10000);		//10ms
					
					can_amount = AZ_VC_OBD2_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{

						if(AZ_VC_OBD2_Read(&msg, &read_size) == 0)
						{
							fprintf(stdout, "OBD2 read size = %d", read_size);
							fprintf(stdout, "\treceive cnt = %d\n",rec_cnt + 1);
							fprintf(stdout, "port = %d", msg.port);
							if(msg.can_mode == 0)
							{
								fprintf(stdout, "\tID = 0x%03X", msg.id);
							}
							else
							{
								fprintf(stdout, "\tID = 0x%08X", msg.id);
							}
							fprintf(stdout, "\tservice = 0x%02X", msg.service);
							fprintf(stdout, "\tpid = 0x%02X", msg.pid);
							fprintf(stdout, "\tcan_mode = 0x%02X", msg.can_mode);
							fprintf(stdout, "\tdlc = 0x%02X\n", msg.dlc);
							for(k=0; k<8; k++) {
								fprintf(stdout, "data[%d] = 0x%02X\t", k, msg.data[k]);
							}
							fprintf(stdout, "\n");
							rec_cnt ++;
							usleep(1);
							break;
						}
					}
					usleep(50000);
					fprintf(stdout,"\n");
				}
			}
			break;

		case OBD2_TEST_IDX_WRITE_sample_PID:
			{
				int i, j, k, rec_cnt;
				uint32_t read_size = 0;
				/*
				uint8_t Service1_2_PidLen[95] = {
					0, 4, 2, 2, 1, 1, 1, 1, 1, 1, 1,  // PID 1 ~ A (0 reserved)
					1, 2, 1, 1, 1, 2, 1, 1, 1, 2,     // PID B ~ 14
				};
				*/
				uint8_t Service1_2_PidLen[16] = {
					1, 2, 1, 1, 2,  // PID 0D,0C,2F,46,5E
					2, 1, 1, 2, 1,  // PID 42,04,05,10,45
					1, 1, 1, 1, 1,  // PID 33,0F,0A,5C,5A
					8,			//PID 2C,
				};
				obd2_send_msg.can_mode = 0;
				obd2_send_msg.dlc = 8;
				obd2_send_msg.id = ntohl(0x000007DF);
				obd2_send_msg.data[0] = 0x02;
				if(obd2_test_sample_cnt != 16)
					obd2_send_msg.data[1] = 0x01;
				else
					obd2_send_msg.data[1] = 0x09;
				//fprintf(stdout,"obd2_test_cnt = %d\n",obd2_test_sample_cnt);
				switch(obd2_test_sample_cnt)
				{
					case 1:	obd2_send_msg.PID = 0x0D;    break;
					case 2:	obd2_send_msg.PID = 0x0C; 	break;
					case 3:	obd2_send_msg.PID = 0x2F;	break;
					case 4:	obd2_send_msg.PID = 0x46;	break;
					case 5:	obd2_send_msg.PID = 0x5E;	break;
					case 6:	obd2_send_msg.PID = 0x42;	break;
					case 7:	obd2_send_msg.PID = 0x04;	break;
					case 8:	obd2_send_msg.PID = 0x05;	break;
					case 9:	obd2_send_msg.PID = 0x10;	break;
					case 10: obd2_send_msg.PID = 0x45;	break;
					case 11:	obd2_send_msg.PID = 0x33;	break;
					case 12:	obd2_send_msg.PID = 0x0F;	break;
					case 13:	obd2_send_msg.PID = 0x0A;	break;
					case 14:	obd2_send_msg.PID = 0x5C;	break;
					case 15:	obd2_send_msg.PID = 0x5A;	break;
					case 16:	obd2_send_msg.PID = 0x02;	break;					
					default:	
						fprintf(stdout, "Select item is not in PID select list\n");
						goto err;
						//obd2_send_msg.PID = 0xFF;	
						break;
				}
				fprintf(stdout,"obd2 send ID = 0x%02X\n",obd2_send_msg.PID);
				obd2_send_msg.data[2] = obd2_send_msg.PID;				
				for(i = 0; i < test_cnt; i++)
				{
					ret = AZ_VC_OBD2_Write(obd2_send_msg);
					if(ret == -1) {
						fprintf(stderr, "Calling AZ_VC_OBD2_Write fail\n");
						goto err;
					}

					usleep(100000);		//100ms
					can_amount = AZ_VC_OBD2_Read_Amount();
					//fprintf(stdout, "ret = %d\n",ret);					
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						if(AZ_VC_OBD2_Read(&msg, &read_size) == 0)
						{
							fprintf(stdout, "receive cnt = %d\n",rec_cnt + 1);
							if(msg.can_mode == 0)
							{
								fprintf(stdout, "\tID = 0x%03X\n", msg.id);
							}
							else
							{
								fprintf(stdout, "\tID = 0x%08X\n", msg.id);
							}
							fprintf(stdout, "\tservice = 0x%02X\n", msg.service);
							fprintf(stdout, "\tpid = 0x%02X\n", msg.pid);
							fprintf(stdout, "\tdata = ");
							for(k=0; k<Service1_2_PidLen[obd2_test_sample_cnt - 1]; k++) {
								fprintf(stdout, "0x%02X ", msg.data[k]);
							}
							fprintf(stdout, "\n");
							rec_cnt ++;
							//usleep(1);
						}
					}
					fprintf(stdout,"\n");
					sleep(1);
				}
			}
			break;


		case OBD2_TEST_IDX_WRITE_sample_PID_DECODE:
			{
				int i, j, k, rec_cnt;
				uint32_t read_size = 0;
				/*
				uint8_t Service1_2_PidLen[95] = {
					0, 4, 2, 2, 1, 1, 1, 1, 1, 1, 1,  // PID 1 ~ A (0 reserved)
					1, 2, 1, 1, 1, 2, 1, 1, 1, 2,     // PID B ~ 14
				};
				*/
				uint8_t Service1_2_PidLen[16] = {
					1, 2, 1, 1, 2,  // PID 0D,0C,2F,46,5E
					2, 1, 1, 2, 1,  // PID 42,04,05,10,45
					1, 1, 1, 1, 1,  // PID 33,0F,0A,5C,5A
					8,		     // PID 2C
				};
				obd2_send_msg.can_mode = 0;
				obd2_send_msg.dlc = 8;
				obd2_send_msg.id = ntohl(0x000007DF);
				obd2_send_msg.data[0] = 0x02;
				if(obd2_test_sample_cnt != 16)
					obd2_send_msg.data[1] = 0x01;
				else
					obd2_send_msg.data[1] = 0x09;
				//fprintf(stdout,"obd2_test_cnt = %d\n",obd2_test_sample_cnt);
				switch(obd2_test_sample_cnt)
				{
					case 1:	obd2_send_msg.PID = 0x0D;    break;
					case 2:	obd2_send_msg.PID = 0x0C; 	break;
					case 3:	obd2_send_msg.PID = 0x2F;	break;
					case 4:	obd2_send_msg.PID = 0x46;	break;
					case 5:	obd2_send_msg.PID = 0x5E;	break;
					case 6:	obd2_send_msg.PID = 0x42;	break;
					case 7:	obd2_send_msg.PID = 0x04;	break;
					case 8:	obd2_send_msg.PID = 0x05;	break;
					case 9:	obd2_send_msg.PID = 0x10;	break;
					case 10: obd2_send_msg.PID = 0x45;	break;
					case 11:	obd2_send_msg.PID = 0x33;	break;
					case 12:	obd2_send_msg.PID = 0x0F;	break;
					case 13:	obd2_send_msg.PID = 0x0A;	break;
					case 14:	obd2_send_msg.PID = 0x5C;	break;
					case 15:	obd2_send_msg.PID = 0x5A;	break;
					case 16:	obd2_send_msg.PID = 0x02;	break;					
					default:	
						fprintf(stdout, "Select item is not in PID select list\n");
						goto err;
						//obd2_send_msg.PID = 0xFF;	
						break;
				}
				fprintf(stdout,"obd2 send ID = 0x%02X\n",obd2_send_msg.PID);
				obd2_send_msg.data[2] = obd2_send_msg.PID;				
				for(i = 0; i < test_cnt; i++)
				{
					if(stop_thread != 0)
						break;
					ret = AZ_VC_OBD2_Write(obd2_send_msg);
					if(ret == -1) {
						fprintf(stderr, "Calling AZ_VC_OBD2_Write fail\n");
						goto err;
					}

					usleep(100000);		//100ms
					struct OBD2_Decode_message_t  decode_msg;
					can_amount = AZ_VC_OBD2_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						if(AZ_VC_OBD2_Read_Decode(&decode_msg, obd2_send_msg.PID, &read_size) == 0)
						{
							fprintf(stdout, "receive cnt = %d\n",rec_cnt + 1);
							fprintf(stdout, "\tpid = 0x%02X\n", decode_msg.pid);
						
							switch(decode_msg.pid)
							{
								//Engine RPM
								case 0x0C:	fprintf(stdout, "\tRPM = %d ", decode_msg.data);	break;

								//vehicle speed(VSS)
								case 0x0D:	fprintf(stdout, "\tSpeed = %d km", decode_msg.data);	break;	

								//Fuel Tank Level Input(FLI)
								case 0x2F:	fprintf(stdout, "\tFLI = %d.%2d ", decode_msg.data,decode_msg.data_f);	break;

								//Ambient air temperature(AAT)
								case 0x46:	fprintf(stdout, "\tAAT = %d %%", decode_msg.data);		break;

								//Engine fuel rate(EFR)
								case 0x5E:	fprintf(stdout, "\tEFR = %d L/h", decode_msg.data);	break;
									
								//Control module voltage(VEP)
								case 0x42:	fprintf(stdout, "\tVEP = %d.%d v", decode_msg.data,decode_msg.data_f);	break;

								//Calculated engine load(PCT)
								case 0x04:	fprintf(stdout, "\tPCT = %d %%", decode_msg.data);	break;

								//Engine coolant temperature(ECT)
								case 0x05:	fprintf(stdout, "\tECT = %d %%", decode_msg.data);		break;

								//Mass air flow sensor(MAF)
								case 0x10:	fprintf(stdout, "\tMAF = %d.%d grams/sec", decode_msg.data,decode_msg.data_f);		break;

								//Relative throttle position(ETP)
								case 0x45:	fprintf(stdout, "\tETP = %d %%", decode_msg.data);		break;

								//Absolute Barometric Pressure(ABP)
								case 0x33:	fprintf(stdout, "\tABP = %d kpa", decode_msg.data);		break;

								//Intake air temperature(IAT)
								case 0x0F:	fprintf(stdout, "\tIAT = %d.C", decode_msg.data);		break;
									
								//Fuel pressure  Pressure(FRP)								
								case 0x0A:	fprintf(stdout, "\tFRP = %d kpa", decode_msg.data);		break;	

								//Engine oil temperature(EOT)
								case 0x5C:	fprintf(stdout, "\tEOT = %d .c", decode_msg.data);		break;	

								//Relative accelerator pedal position(EEC2)
								case 0x5A:	fprintf(stdout, "\tEEC2 = %d %%", decode_msg.data);		break;

								//Relative vehicle identification number(VIN)
								case 0x02:	
									fprintf(stdout,"\tVIN = ");
									for(k =0; k < sizeof(decode_msg.data_vi); k++)
										fprintf(stdout, "%02X", decode_msg.data_vi[k]);		
									break;
							}							
							fprintf(stdout, "\n");
							rec_cnt ++;
							//usleep(1);
						}
					}
					fprintf(stdout,"\n");
					
					sleep(1);
				}
			}
			break;
			
case OBD2_TEST_IDX_WRITE_sample_PID_EXT:
			{
				int i, j, k, rec_cnt;
				uint32_t read_size = 0;
				/*
				uint8_t Service1_2_PidLen[95] = {
					0, 4, 2, 2, 1, 1, 1, 1, 1, 1, 1,  // PID 1 ~ A (0 reserved)
					1, 2, 1, 1, 1, 2, 1, 1, 1, 2,     // PID B ~ 14
				};
				*/
				uint8_t Service1_2_PidLen[16] = {
					1, 2, 1, 1, 2,  // PID 0D,0C,2F,46,5E
					2, 1, 1, 2, 1,  // PID 42,04,05,10,45
					1, 1, 1, 1, 1,  // PID 33,0F,0A,5C,5A
					8,			//PID 2C,
				};
				obd2_send_msg.can_mode = 1;
				obd2_send_msg.dlc = 8;
				obd2_send_msg.id = ntohl(0x18DB33F1);
				obd2_send_msg.data[0] = 0x02;
				if(obd2_test_sample_cnt != 16)
					obd2_send_msg.data[1] = 0x01;
				else
					obd2_send_msg.data[1] = 0x09;
				//fprintf(stdout,"obd2_test_cnt = %d\n",obd2_test_sample_cnt);
				switch(obd2_test_sample_cnt)
				{
					case 1:	obd2_send_msg.PID = 0x0D;    break;
					case 2:	obd2_send_msg.PID = 0x0C; 	break;
					case 3:	obd2_send_msg.PID = 0x2F;	break;
					case 4:	obd2_send_msg.PID = 0x46;	break;
					case 5:	obd2_send_msg.PID = 0x5E;	break;
					case 6:	obd2_send_msg.PID = 0x42;	break;
					case 7:	obd2_send_msg.PID = 0x04;	break;
					case 8:	obd2_send_msg.PID = 0x05;	break;
					case 9:	obd2_send_msg.PID = 0x10;	break;
					case 10: obd2_send_msg.PID = 0x45;	break;
					case 11:	obd2_send_msg.PID = 0x33;	break;
					case 12:	obd2_send_msg.PID = 0x0F;	break;
					case 13:	obd2_send_msg.PID = 0x0A;	break;
					case 14:	obd2_send_msg.PID = 0x5C;	break;
					case 15:	obd2_send_msg.PID = 0x5A;	break;
					case 16:	obd2_send_msg.PID = 0x02;	break;					
					default:	
						fprintf(stdout, "Select item is not in PID select list\n");
						goto err;
						//obd2_send_msg.PID = 0xFF;	
						break;
				}
				fprintf(stdout,"obd2 send ID = 0x%02X\n",obd2_send_msg.PID);
				obd2_send_msg.data[2] = obd2_send_msg.PID;				
				for(i = 0; i < test_cnt; i++)
				{
					ret = AZ_VC_OBD2_Write(obd2_send_msg);
					if(ret == -1) {
						fprintf(stderr, "Calling AZ_VC_OBD2_Write fail\n");
						goto err;
					}

					usleep(100000);		//100ms
					can_amount = AZ_VC_OBD2_Read_Amount();
					//fprintf(stdout, "ret = %d\n",ret);					
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						if(AZ_VC_OBD2_Read(&msg, &read_size) == 0)
						{
							fprintf(stdout, "receive cnt = %d\n",rec_cnt + 1);
							fprintf(stdout, "\tid = 0x%02X\n", msg.id);
							fprintf(stdout, "\tservice = 0x%02X\n", msg.service);
							fprintf(stdout, "\tpid = 0x%02X\n", msg.pid);
							fprintf(stdout, "\tdata = ");
							for(k=0; k<Service1_2_PidLen[obd2_test_sample_cnt - 1]; k++) {
								fprintf(stdout, "0x%02X ", msg.data[k]);
							}
							fprintf(stdout, "\n");
							rec_cnt ++;
							//usleep(1);
						}
					}
					fprintf(stdout,"\n");
					sleep(1);
				}
			}
			break;


		case OBD2_TEST_IDX_WRITE_sample_PID_DECODE_EXT:
			{
				int i, j, k, rec_cnt;
				uint32_t read_size = 0;
				/*
				uint8_t Service1_2_PidLen[95] = {
					0, 4, 2, 2, 1, 1, 1, 1, 1, 1, 1,  // PID 1 ~ A (0 reserved)
					1, 2, 1, 1, 1, 2, 1, 1, 1, 2,     // PID B ~ 14
				};
				*/
				uint8_t Service1_2_PidLen[16] = {
					1, 2, 1, 1, 2,  // PID 0D,0C,2F,46,5E
					2, 1, 1, 2, 1,  // PID 42,04,05,10,45
					1, 1, 1, 1, 1,  // PID 33,0F,0A,5C,5A
					8,		     // PID 2C
				};
				obd2_send_msg.can_mode = 1;
				obd2_send_msg.dlc = 8;
				obd2_send_msg.id = ntohl(0x18DB33F1);
				obd2_send_msg.data[0] = 0x02;
				if(obd2_test_sample_cnt != 16)
					obd2_send_msg.data[1] = 0x01;
				else
					obd2_send_msg.data[1] = 0x09;
				//fprintf(stdout,"obd2_test_cnt = %d\n",obd2_test_sample_cnt);
				switch(obd2_test_sample_cnt)
				{
					case 1:	obd2_send_msg.PID = 0x0D;    break;
					case 2:	obd2_send_msg.PID = 0x0C; 	break;
					case 3:	obd2_send_msg.PID = 0x2F;	break;
					case 4:	obd2_send_msg.PID = 0x46;	break;
					case 5:	obd2_send_msg.PID = 0x5E;	break;
					case 6:	obd2_send_msg.PID = 0x42;	break;
					case 7:	obd2_send_msg.PID = 0x04;	break;
					case 8:	obd2_send_msg.PID = 0x05;	break;
					case 9:	obd2_send_msg.PID = 0x10;	break;
					case 10: obd2_send_msg.PID = 0x45;	break;
					case 11:	obd2_send_msg.PID = 0x33;	break;
					case 12:	obd2_send_msg.PID = 0x0F;	break;
					case 13:	obd2_send_msg.PID = 0x0A;	break;
					case 14:	obd2_send_msg.PID = 0x5C;	break;
					case 15:	obd2_send_msg.PID = 0x5A;	break;
					case 16:	obd2_send_msg.PID = 0x02;	break;					
					default:	
						fprintf(stdout, "Select item is not in PID select list\n");
						goto err;
						//obd2_send_msg.PID = 0xFF;	
						break;
				}
				fprintf(stdout,"obd2 send ID = 0x%02X\n",obd2_send_msg.PID);
				obd2_send_msg.data[2] = obd2_send_msg.PID;				
				for(i = 0; i < test_cnt; i++)
				{
					if(stop_thread != 0)
						break;
					ret = AZ_VC_OBD2_Write(obd2_send_msg);
					if(ret == -1) {
						fprintf(stderr, "Calling AZ_VC_OBD2_Write fail\n");
						goto err;
					}

					usleep(100000);		//100ms
					struct OBD2_Decode_message_t  decode_msg;
					can_amount = AZ_VC_OBD2_Read_Amount();
					for(can_index = 0; can_index < can_amount; can_index++)
					{
						if(AZ_VC_OBD2_Read_Decode(&decode_msg, obd2_send_msg.PID, &read_size) == 0)
						{
							fprintf(stdout, "receive cnt = %d\n",rec_cnt + 1);
							fprintf(stdout, "\tpid = 0x%02X\n", decode_msg.pid);
						
							switch(decode_msg.pid)
							{
								//Engine RPM
								case 0x0C:	fprintf(stdout, "\tRPM = %d ", decode_msg.data);	break;

								//vehicle speed(VSS)
								case 0x0D:	fprintf(stdout, "\tSpeed = %d km", decode_msg.data);	break;	

								//Fuel Tank Level Input(FLI)
								case 0x2F:	fprintf(stdout, "\tFLI = %d.%2d ", decode_msg.data,decode_msg.data_f);	break;

								//Ambient air temperature(AAT)
								case 0x46:	fprintf(stdout, "\tAAT = %d %%", decode_msg.data);		break;

								//Engine fuel rate(EFR)
								case 0x5E:	fprintf(stdout, "\tEFR = %d L/h", decode_msg.data);	break;
									
								//Control module voltage(VEP)
								case 0x42:	fprintf(stdout, "\tVEP = %d.%d v", decode_msg.data,decode_msg.data_f);	break;

								//Calculated engine load(PCT)
								case 0x04:	fprintf(stdout, "\tPCT = %d %%", decode_msg.data);	break;

								//Engine coolant temperature(ECT)
								case 0x05:	fprintf(stdout, "\tECT = %d %%", decode_msg.data);		break;

								//Mass air flow sensor(MAF)
								case 0x10:	fprintf(stdout, "\tMAF = %d.%d grams/sec", decode_msg.data,decode_msg.data_f);		break;

								//Relative throttle position(ETP)
								case 0x45:	fprintf(stdout, "\tETP = %d %%", decode_msg.data);		break;

								//Absolute Barometric Pressure(ABP)
								case 0x33:	fprintf(stdout, "\tABP = %d kpa", decode_msg.data);		break;

								//Intake air temperature(IAT)
								case 0x0F:	fprintf(stdout, "\tIAT = %d.C", decode_msg.data);		break;
									
								//Fuel pressure  Pressure(FRP)								
								case 0x0A:	fprintf(stdout, "\tFRP = %d kpa", decode_msg.data);		break;	

								//Engine oil temperature(EOT)
								case 0x5C:	fprintf(stdout, "\tEOT = %d .c", decode_msg.data);		break;	

								//Relative accelerator pedal position(EEC2)
								case 0x5A:	fprintf(stdout, "\tEEC2 = %d %%", decode_msg.data);		break;

								//Relative vehicle identification number(VIN)
								case 0x02:	
									fprintf(stdout,"\tVIN = ");
									for(k =0; k < sizeof(decode_msg.data_vi); k++)
										fprintf(stdout, "%02X", decode_msg.data_vi[k]);		
									break;
							}							
							fprintf(stdout, "\n");
							rec_cnt ++;
							//usleep(1);
						}
					}
					fprintf(stdout,"\n");
					
					sleep(1);
				}
			}
			break;
			
			
		default:
			fprintf(stderr, "Invalid arguments\n");
			goto err;
	}

err:
	pthread_exit(NULL);
}

static int do_functions(void)
{
	int ret = 0;
	pthread_t deinit_wait_thread;
	pthread_t do_get_can_fw_ver_thread;
	pthread_t do_can_config_thread;
	pthread_t do_mode_active_thread;
	pthread_t do_obd2_demo_thread;
	pthread_t do_filter_config_raw_thread;
	pthread_t do_get_filter_config_thread;

	fprintf(stdout, "version %d.%d.%d\n",
		FARO_CAN_SDK_DEMO_VERSION_MAJOR,
		FARO_CAN_SDK_DEMO_VERSION_MINOR,
		FARO_CAN_SDK_DEMO_VERSION_PATCH);

	if(scan_port) {
		if((ret = do_scan_port()) != 0) {
			fprintf(stderr, "do_scan_port fail\n");
		}

		goto err;
	}

	// spawn deinit_wait_thread if necessary
	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "spawning deinit_wait thread...\n");

	if((ret = pthread_create(&deinit_wait_thread, NULL, deinit_wait, NULL)) != 0) {
		fprintf(stderr, "create deinit_wait thread fail\n");
			goto err;
	}

	if(get_can_fw_ver) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_get_can_fw_ver thread...\n");

		if((ret = pthread_create(&do_get_can_fw_ver_thread, NULL, do_get_can_fw_ver, NULL)) != 0) {
			fprintf(stderr, "create do_get_can_fw_ver thread fail\n");
			goto err;
		}

		ret = pthread_join(do_get_can_fw_ver_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_get_can_fw_ver_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(can_config) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_can_config thread...\n");
		if((ret = pthread_create(&do_can_config_thread, NULL, do_can_config, NULL)) != 0) {
			fprintf(stderr, "create do_can_config thread fail\n");
			goto err;
		}

		ret = pthread_join(do_can_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_can_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(mode_active) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_mode_active thread...\n");
		if((ret = pthread_create(&do_mode_active_thread, NULL, do_mode_active, NULL)) != 0) {
			fprintf(stderr, "create do_mode_active thread fail\n");
			goto err;
		}

		ret = pthread_join(do_mode_active_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_mode_active_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(filter_config_raw) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_filter_config_raw thread...\n");
		if((ret = pthread_create(&do_filter_config_raw_thread, NULL, do_filter_config_raw, &can_filter_config)) != 0) {
			fprintf(stderr, "create do_filter_config_raw thread fail\n");
			goto err;
		}

		ret = pthread_join(do_filter_config_raw_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_filter_config_raw_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(get_filter_config) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_get_filter_config thread...\n");
		if((ret = pthread_create(&do_get_filter_config_thread, NULL, do_get_filter_config, NULL)) != 0) {
			fprintf(stderr, "create do_get_filter_config thread fail\n");
			goto err;
		}

		ret = pthread_join(do_get_filter_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_get_filter_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(obd2_test) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn do_obd2_demo thread...\n");
		if((ret = pthread_create(&do_obd2_demo_thread, NULL, do_obd2_demo, NULL)) != 0) {
			fprintf(stderr, "create do_obd2_demo thread fail\n");
			goto err;
		}

		ret = pthread_join(do_obd2_demo_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_obd2_demo error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

normal_exit:
	ret = pthread_join(deinit_wait_thread, NULL);
	if(ret) {
		fprintf(stderr, "deinit_wait thread error: %s\n", strerror(ret));
		goto err;
	}


	fprintf(stderr, "threads are stopped\n");
	fflush(stderr);

err:
	scan_port = 0;
	get_can_fw_ver = 0;
	can_config = 0;
	mode_active = 0;
	obd2_test = 0;
	return ret;
}

int main(int argc, char *argv[])
{
	int ret;

	if(signal(SIGINT, sig_handler) == SIG_ERR)
		fprintf(stderr, "fail to catch SIGINT\n");

	if(process_options(argc, argv) == 0)
		do_functions();

    if(port_name != NULL)
    	free(port_name);

	return 0;
}
